# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Class to represent files."""

from __future__ import annotations

import dataclasses
import os
import warnings
from datetime import datetime, timezone
from pathlib import Path
from typing import NoReturn, Optional, Union, cast

from .error import IntegrityError
from .filesystem import RemotePath, checksum_of_file, file_modification_time, file_size
from .logging import get_logger
from .model import DataFile


@dataclasses.dataclass(frozen=True)
class File:
    """Store local and remote paths and metadata for a file.

    There are two central properties:

    - ``remote_path``: Path to the remote file relative to the dataset's
      ``source_folder``. This is always set, even if the file does not exist
      on the remote filesystem.
    - ``local_path``: Path to the file on the local filesystem.
      Is ``None`` if the file does not exist locally.

    Files can be in one of three states and the state can be changed as shown below.
    The state can be queried using :meth:`File.is_on_local`
    and :meth:`File.is_on_remote`.

    .. code-block:: none

         local                                  remote
           │                                      │
           │ uploaded                  downloaded │
           │                                      │
           └───────────> local+remote <───────────┘
    """

    local_path: Optional[Path]
    """Path to the file on the local filesystem."""
    remote_path: RemotePath
    """Path to the file on the remote filesystem."""
    remote_gid: Optional[str]
    """Unix group ID on remote."""
    remote_perm: Optional[str]
    """Unix file mode on remote."""
    remote_uid: Optional[str]
    """Unix user ID on remote."""
    created_at: Optional[datetime] = None
    """Creator of the file entry in SciCat."""
    created_by: Optional[str] = None
    """Creation time of the file entry in SciCat."""
    updated_at: Optional[datetime] = None
    """Last updator of the file entry in SciCat."""
    updated_by: Optional[str] = None
    """Last update time of the file entry in SciCat."""
    checksum_algorithm: Optional[str] = None
    """Algorithm to use for checksums."""
    _remote_size: Optional[int] = dataclasses.field(default=None, repr=False)
    _remote_creation_time: Optional[datetime] = dataclasses.field(
        default=None, repr=False
    )
    _remote_checksum: Optional[str] = dataclasses.field(default=None, repr=False)
    _checksum_cache: Optional[_Checksum] = dataclasses.field(
        default=None, compare=False, repr=False
    )

    @classmethod
    def from_local(
        cls,
        path: Union[str, Path],
        *,
        base_path: Union[str, Path] = "",
        remote_path: Optional[Union[str, RemotePath]] = None,
        remote_uid: Optional[str] = None,
        remote_gid: Optional[str] = None,
        remote_perm: Optional[str] = None,
    ) -> File:
        """Link a file on the local filesystem.

        Given following ``path``, ``base_path``, and ``source_folder``::

            path:                      somewhere/on/local/folder/file.nxs
            base_bath:                 somewhere/on/local
            source_folder:             remote/storage

        the file will be located on the remote at::

            -> remote_path:            folder/file.nxs
            -> actual remote location: remote/storage/folder/file.nxs

        ``remote_path`` can also be overridden, in which case ``path`` and
        ``base_path`` are not used to deduce the "actual remote location".

        Parameters
        ----------
        path:
            Full path the local file.
        base_path:
            Only use ``path.relative_to(base_path)`` to determine the remote location.
        remote_path:
            Path on the remote, relative to ``source_folder``.
        remote_uid:
            User ID on the remote. Will be determined automatically on upload.
        remote_gid:
            Group ID on the remote. Will be determined automatically on upload.
        remote_perm:
            File permissions on the remote. Will be determined automatically on upload.
        """
        path = Path(path)
        if not remote_path:
            remote_path = str(path.relative_to(base_path))
        return File(
            local_path=path,
            remote_path=RemotePath(remote_path),
            remote_gid=remote_gid,
            remote_perm=remote_perm,
            remote_uid=remote_uid,
            _checksum_cache=_Checksum(),
        )

    @classmethod
    def from_scicat(
        cls,
        model: DataFile,
        *,
        local_path: Optional[Union[str, Path]] = None,
    ) -> File:
        """Construct a new file object from SciCat models.

        Parameters
        ----------
        model:
            Pydantic model for the file.
        local_path:
            Value for the local path.

        Returns
        -------
        :
            A new file object.
        """
        return File(
            local_path=Path(local_path) if isinstance(local_path, str) else local_path,
            remote_path=RemotePath(model.path),
            remote_gid=model.gid,
            remote_perm=model.perm,
            remote_uid=model.uid,
            created_at=model.createdAt,
            created_by=model.createdBy,
            updated_at=model.updatedAt,
            updated_by=model.updatedBy,
            _remote_size=model.size,
            _remote_creation_time=model.time,
            _remote_checksum=model.chk,
        )

    @property
    def size(self) -> int:
        """The size in bytes of the file.

        If the file exists on local, return the current size of the local file.
        Otherwise, return the stored size in the catalogue.
        """
        if self.is_on_local:
            return file_size(cast(Path, self.local_path))
        return self._remote_size  # type: ignore[return-value]

    @property
    def creation_time(self) -> datetime:
        """The logical creation time of the SciCat file.

        If the file exists on local, return the time the local file was last modified.
        Otherwise, return the stored time in the catalogue.
        """
        if self.is_on_local:
            return file_modification_time(cast(Path, self.local_path))
        return self._remote_creation_time  # type: ignore[return-value]

    def checksum(self) -> Optional[str]:
        """Return the checksum of the file.

        This can take a long time to compute for large files.

        If the file exists on local, return the current checksum of the local file.
        Otherwise, return the stored checksum in the catalogue.

        Returns
        -------
        :
            The checksum of the file.
        """
        if not self.is_on_local:
            return self._remote_checksum
        if self.checksum_algorithm is None:
            return None
        return self._checksum_cache.get(  # type: ignore[union-attr]
            path=self.local_path,  # type: ignore[arg-type]
            algorithm=self.checksum_algorithm,
        )

    def remote_access_path(
        self, source_folder: Union[RemotePath, str]
    ) -> Optional[RemotePath]:
        """Full path to the file on the remote if it exists."""
        return (source_folder / self.remote_path) if self.is_on_remote else None

    @property
    def is_on_remote(self) -> bool:
        """True if the file is on remote."""
        return self._remote_size is not None

    @property
    def is_on_local(self) -> bool:
        """True if the file is on local."""
        return self.local_path is not None

    def local_is_up_to_date(self) -> bool:
        """Check if the file on local is up-to-date.

        Returns
        -------
        :
            True if the file exists on local and its checksum
            matches the stored checksum for the remote file.
        """
        if not self.is_on_remote:
            return True
        if not self.is_on_local:
            return False
        if self.checksum_algorithm is None:
            warnings.warn(
                f"Cannot check if local file {self.local_path} is up to date because "
                "the checksum algorithm is not set. "
                "Assuming the file needs to be updated."
            )
            return False
        local_checksum = self._checksum_cache.get(  # type: ignore[union-attr]
            path=self.local_path,  # type: ignore[arg-type]
            algorithm=self.checksum_algorithm,
        )
        return self._remote_checksum == local_checksum

    def make_model(self, *, for_archive: bool = False) -> DataFile:
        """Build a pydantic model for this file.

        Parameters
        ----------
        for_archive:
            Select whether the file is stored in an archive or on regular disk,
            that is whether it belongs to a Datablock or an OrigDatablock.

        Returns
        -------
        :
            A new pydantic model.
        """
        chk = self.checksum()
        # TODO if for_archive: ensure not out of date
        return DataFile(
            path=os.fspath(self.remote_path),
            size=self.size,
            chk=chk,
            gid=self.remote_gid,
            perm=self.remote_perm,
            time=self.creation_time,
            uid=self.remote_uid,
        )

    def uploaded(
        self,
        *,
        remote_path: Optional[Union[str, RemotePath]] = None,
        remote_uid: Optional[str] = None,
        remote_gid: Optional[str] = None,
        remote_perm: Optional[str] = None,
        remote_creation_time: Optional[datetime] = None,
        remote_size: Optional[int] = None,
    ) -> File:
        """Return new file metadata after an upload.

        Assumes that the input file exists on local.
        The returned object is on both local and remote.

        Parameters
        ----------
        remote_path:
            New remote path.
        remote_uid:
            New user ID on remote, overwrites any current value.
        remote_gid:
            New group ID on remote, overwrites any current value.
        remote_perm:
            New unix permissions on remote, overwrites any current value.
        remote_creation_time:
            Time the file became available on remote.
            Defaults to the current time in UTC.
        remote_size:
            File size on remote.

        Returns
        -------
        :
            A new file object.
        """
        if remote_creation_time is None:
            remote_creation_time = datetime.now().astimezone(timezone.utc)
        args = dict(
            remote_path=RemotePath(remote_path) if remote_path is not None else None,
            remote_gid=remote_gid,
            remote_uid=remote_uid,
            remote_perm=remote_perm,
            _remote_creation_time=remote_creation_time,
        )
        return dataclasses.replace(
            self,
            _remote_size=remote_size if remote_size is not None else self.size,
            _remote_checksum=self.checksum(),
            **{key: val for key, val in args.items() if val is not None},
        )

    def downloaded(self, *, local_path: Union[str, Path]) -> File:
        """Return new file metadata after a download.

        Assumes that the input file exists on remote.
        The returned object is on both local and remote.

        Parameters
        ----------
        local_path:
            New local path.

        Returns
        -------
        :
            A new file object.
        """
        return dataclasses.replace(
            self, local_path=Path(local_path), _checksum_cache=_Checksum()
        )

    def validate_after_download(self) -> None:
        """Check that the file on disk matches the metadata.

        Compares file size and, if possible, its checksum.
        Raises on failure.
        If the function returns without exception, the file is valid.

        Raises
        ------
        IntegrityError
            If a check fails.
        """
        self._validate_after_download_file_size()
        if self._remote_checksum is None:
            get_logger().info(
                "Dataset does not contain a checksum for file '%s'. Skipping check.",
                self.local_path,
            )
            return
        stored = self._remote_checksum
        if not self.checksum_algorithm:
            get_logger().warning(
                "File '%s' has a checksum but no algorithm has been set. "
                "Skipping check. Checksum is %s",
                self.local_path,
                stored,
            )
            return
        actual = checksum_of_file(
            cast(Path, self.local_path), algorithm=self.checksum_algorithm
        )
        if actual != stored:
            _log_and_raise(
                IntegrityError,
                f"Checksum of file '{self.local_path}' ({actual}) "
                f"does not match checksum stored in dataset "
                f"({stored}). Using algorithm "
                f"'{self.checksum_algorithm}'.",
            )

    def _validate_after_download_file_size(self) -> None:
        actual = file_size(cast(Path, self.local_path))
        if actual != self._remote_size:
            get_logger().info(
                "Size of downloaded file '%s' (%d bytes) does not "
                "match size reported in dataset (%d bytes)."
                "This may be due to a difference in file systems and perfectly fine. "
                "Or it is caused by an error during download.",
                self.local_path,
                actual,
                self._remote_size,
            )


def _log_and_raise(typ: type, msg: str) -> NoReturn:
    get_logger().error(msg)
    raise typ(msg)


class _Checksum:
    """Compute and cache the checksum of a file."""

    def __init__(self) -> None:
        self._value: Optional[str] = None
        self._path: Optional[Path] = None
        self._algorithm: Optional[str] = None
        self._access_time: Optional[datetime] = None

    def get(self, *, path: Path, algorithm: str) -> str:
        if self._is_out_of_date(path=path, algorithm=algorithm):
            self._update(path=path, algorithm=algorithm)
        return self._value  # type: ignore[return-value]

    def _is_out_of_date(self, *, path: Path, algorithm: str) -> bool:
        return (
            self._access_time is None
            or path != self._path
            or algorithm != self._algorithm
            or file_modification_time(path) > self._access_time
        )

    def _update(self, *, path: Path, algorithm: str) -> None:
        self._value = checksum_of_file(path, algorithm=algorithm)
        self._path = path
        self._algorithm = algorithm
        self._access_time = datetime.now(tz=timezone.utc)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""HTML representations for Jupyter."""

import html
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional

import pydantic

from ..dataset import Dataset
from ..filesystem import RemotePath
from ..model import DatasetLifecycle, DatasetType, Technique
from ..pid import PID
from . import resources


def dataset_html_repr(dset: Dataset) -> str:
    template = resources.dataset_repr_template()
    style_sheet = resources.dataset_style()
    fields = _get_fields(dset)
    main_rows = "\n".join(_format_field(field) for field in fields if field.main)
    detail_rows = "\n".join(_format_field(field) for field in fields if not field.main)
    return template.substitute(
        style_sheet=style_sheet,
        dataset_type=_format_dataset_type(dset.type),
        file_info=_format_file_info(dset, archived=False),
        archived_file_info=_format_file_info(dset, archived=True),
        main_rows=main_rows,
        detail_rows=detail_rows,
        file_rows=_format_files(dset),
        metadata_rows=_format_metadata(dset),
    )


def _format_files(dset: Dataset) -> str:
    template = resources.files_repr_template()
    return "\n".join(
        template.substitute(
            local_path=(
                '<span class="cean-empty-field">None</span>'
                if not file.is_on_local
                else html.escape(str(file.local_path))
            ),
            remote_path=(
                '<span class="cean-empty-field">None</span>'
                if not file.is_on_remote
                else html.escape(str(file.remote_path))
            ),
            size=_human_readable_size(file.size),
        )
        for file in dset.files
    )


def _format_metadata(dset: Dataset) -> str:
    template = resources.metadata_template()
    return "\n".join(
        template.substitute(
            name=html.escape(str(name)),
            value=html.escape(str(value)),
        )
        for name, value in dset.meta.items()
    )


@dataclass
class Field:
    name: str
    value: Any
    type: type
    description: str
    read_only: bool
    required: bool
    error: Optional[str]
    main: bool


def _format_field(field: Field) -> str:
    def format_value(val) -> str:
        if isinstance(val, datetime):
            return val.strftime("%Y-%m-%d %H:%M:%S%z")
        return html.escape(str(val))

    name = field.name
    flag = _format_field_flag(field)
    value = (
        '<span class="cean-empty-field">None</span>'
        if field.value is None
        else format_value(field.value)
    )
    typ = _format_type(field.type)
    description = html.escape(field.description)
    row_highlight = _row_highlight_classes(field)

    template = resources.dataset_field_repr_template()
    return template.substitute(
        name=name,
        flag=flag,
        type=typ,
        value=value,
        description=description,
        extra_classes=row_highlight,
    )


_EXCLUDED_FIELDS = {
    "type",
    "number_of_files",
    "number_of_files_archived",
    "size",
    "packed_size",
    "meta",
}

_MAIN_FIELDS = {
    "pid",
    "source_folder",
    "name",
    "description",
    "creation_time",
    "proposal_id",
    "sample_id",
    "input_datasets",
}


def _get_fields(dset: Dataset) -> List[Field]:
    validation = _validate(dset)
    fields = [
        Field(
            name=field.name,
            value=getattr(dset, field.name, None),
            type=field.type,
            description=field.description,
            read_only=field.read_only,
            required=field.required(dset.type),
            error=_check_error(field, validation),
            main=field.name in _MAIN_FIELDS,
        )
        for field in dset.fields()
        if field.name not in _EXCLUDED_FIELDS
        and (field.used_by(dset.type) or getattr(dset, field.name, None) is not None)
    ]
    return sorted(
        sorted(fields, key=lambda field: field.name),
        key=lambda field: not field.required,
    )


def _check_error(field: Field, validation: Dict[str, str]) -> Optional[str]:
    if field.name in validation:
        # TODO validation uses model names (camelCase)
        return validation[field.name]
    return None


def _validate(dset: Dataset) -> Dict[str, str]:
    def single_elem(xs):
        (x,) = xs
        return x

    try:
        dset.validate()
        return {}
    except pydantic.ValidationError as e:
        return {single_elem(error["loc"]): error["msg"] for error in e.errors()}


_TYPE_NAME = {
    str: "str",
    int: "int",
    bool: "bool",
    datetime: "datetime",
    PID: "PID",
    DatasetLifecycle: "DatasetLifecycle",
    RemotePath: "RemotePath",
    List[str]: "list[str]",
    List[PID]: "list[PID]",
    List[Technique]: "list[Technique]",
    List[dict]: "list[dict]",  # type: ignore[type-arg]
    dict: "dict",
}


def _format_type(typ: Any) -> str:
    try:
        return _TYPE_NAME[typ]
    except KeyError:
        return html.escape(str(typ))


def _format_field_flag(field: Field) -> str:
    flags = ""
    if field.read_only:
        flags += f'<div class="cean-lock">{resources.image("lock.svg")}</div>'
    if field.required:
        flags += '<div style="color: var(--jp-error-color0)">*</div>'
    return flags


def _format_dataset_type(dataset_type: DatasetType) -> str:
    return "Raw" if dataset_type == DatasetType.RAW else "Derived"


def _format_file_info(dset: Dataset, archived: bool) -> str:
    if archived:
        n_files = dset.number_of_files_archived
        if n_files == 0:
            # Don't show archived files if there aren't any.
            # But always show regular files.
            return ""
        size = dset.packed_size
        name = "archived: "
    else:
        n_files = dset.number_of_files
        size = dset.size
        name = ""

    return (
        f'{name} {n_files} <span class="cean-file-info-size">'
        f"({_human_readable_size(size)})</span>"
    )


def _human_readable_size(size_in_bytes: int) -> str:
    for power, prefix in ((4, "T"), (3, "G"), (2, "M"), (1, "k")):
        n = 1024**power
        if size_in_bytes >= n:
            return f"{size_in_bytes/n:.2f} {prefix}iB"
    return f"{size_in_bytes} B"


def _row_highlight_classes(field: Field) -> str:
    if field.required and field.value is None:
        return "cean-missing-value"
    if field.error:
        return "cean-error"
    return ""



##########################################
# This file was automatically generated. #
# Do not modify it directly!             #
##########################################
# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
# flake8: noqa

"""Base dataclass for Dataset."""

from __future__ import annotations

import dataclasses
from datetime import datetime, timezone
from typing import Any, Callable, Dict, List, Literal, Optional, Union

import dateutil.parser

from .datablock import OrigDatablockProxy
from .filesystem import RemotePath
from .model import (
    DatasetLifecycle,
    DatasetType,
    DerivedDataset,
    OrigDatablock,
    RawDataset,
    Technique,
)
from .pid import PID


def _apply_default(
    value: Any, default: Any, default_factory: Optional[Callable[[], Any]]
) -> Any:
    if value is not None:
        return value
    if default_factory is not None:
        return default_factory()
    return default


def _parse_datetime(x: Optional[Union[datetime, str]]) -> datetime:
    if isinstance(x, datetime):
        return x
    if isinstance(x, str):
        if x != "now":
            return dateutil.parser.parse(x)
    return datetime.now(tz=timezone.utc)


class DatasetFields:
    @dataclasses.dataclass(frozen=True)
    class Field:
        name: str
        description: str
        read_only: bool
        required_by_derived: bool
        required_by_raw: bool
        type: type
        used_by_derived: bool
        used_by_raw: bool

        def required(self, dataset_type: DatasetType) -> bool:
            return (
                self.required_by_raw
                if dataset_type == DatasetType.RAW
                else self.required_by_derived
            )

        def used_by(self, dataset_type: DatasetType) -> bool:
            return (
                self.used_by_raw
                if dataset_type == DatasetType.RAW
                else self.used_by_derived
            )

    _FIELD_SPEC = [
        Field(
            name="access_groups",
            description="Groups which have read access to the data. The special group 'public' makes data available to all users.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=List[str],
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="classification",
            description="ACIA information about AUthenticity,COnfidentiality,INtegrity and AVailability requirements of dataset. E.g. AV(ailabilty)=medium could trigger the creation of two tape copies. Format 'AV=medium,CO=low'",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="contact_email",
            description="Email of contact person for this dataset. May contain a list of emails, which should then be separated by semicolons.",
            read_only=False,
            required_by_derived=True,
            required_by_raw=True,
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="created_at",
            description="Time when the object was created in the database.",
            read_only=True,
            required_by_derived=False,
            required_by_raw=False,
            type=datetime,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="created_by",
            description="Account name who created the object in the database.",
            read_only=True,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="creation_location",
            description="Unique location identifier where data was taken, usually in the form /Site-name/facility-name/instrumentOrBeamline-name",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=False,
            used_by_raw=True,
        ),
        Field(
            name="creation_time",
            description="Time when dataset became fully available on disk, i.e. all containing files have been written.",
            read_only=False,
            required_by_derived=True,
            required_by_raw=True,
            type=datetime,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="data_format",
            description="Defines format of subsequent scientific meta data, e.g Nexus Version x.y",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=False,
            used_by_raw=True,
        ),
        Field(
            name="description",
            description="Free text explanation of the contents of the dataset.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="end_time",
            description="Time of end of data taking for this dataset.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=datetime,
            used_by_derived=False,
            used_by_raw=True,
        ),
        Field(
            name="history",
            description="List of previous versions of the dataset. Populated automatically by SciCat.",
            read_only=True,
            required_by_derived=False,
            required_by_raw=False,
            type=List[dict],
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="input_datasets",
            description="Array of input dataset identifiers used in producing the derived dataset. Ideally these are the global identifier to existing datasets inside this or federated data catalogs.",
            read_only=False,
            required_by_derived=True,
            required_by_raw=True,
            type=List[PID],
            used_by_derived=True,
            used_by_raw=False,
        ),
        Field(
            name="instrument_group",
            description="Groups which have read and write access to the data.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="instrument_id",
            description="SciCat ID of the instrument used to measure the data.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=False,
            used_by_raw=True,
        ),
        Field(
            name="investigator",
            description="Name(s) of the investigator(s). The string may contain a list of names, which should then be separated by semicolons.",
            read_only=False,
            required_by_derived=True,
            required_by_raw=True,
            type=str,
            used_by_derived=True,
            used_by_raw=False,
        ),
        Field(
            name="is_published",
            description="Indicate whether the dataset is publicly available.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=bool,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="job_log_data",
            description="The output job logfile. Keep the size of this log data well below 15 MB.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=True,
            used_by_raw=False,
        ),
        Field(
            name="job_parameters",
            description="Input parameters to the job that produced the derived data.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=dict,
            used_by_derived=True,
            used_by_raw=False,
        ),
        Field(
            name="keywords",
            description="Array of tags associated with the meaning or contents of this dataset. Values should ideally come from defined vocabularies, taxonomies, ontologies or knowledge graphs.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=List[str],
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="license",
            description="Name of license under which data can be used.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="lifecycle",
            description="Parameters for storage and publishing of dataset.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=DatasetLifecycle,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="meta",
            description="Free form meta data.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=Dict,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="name",
            description="A name for the dataset.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="number_of_files",
            description="Total number of files in directly accessible storage associated with the dataset. (Corresponds to OrigDatablocks.)",
            read_only=True,
            required_by_derived=False,
            required_by_raw=False,
            type=int,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="number_of_files_archived",
            description="Total number of archived files associated with the dataset. (Corresponds to Datablocks.)",
            read_only=True,
            required_by_derived=False,
            required_by_raw=False,
            type=int,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="orcid_of_owner",
            description="ORCID of owner/custodian. The string may contain a list of ORCID, which should then be separated by semicolons. ORCIDs must include the prefix https://orcid.org/",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="owner",
            description="Owner or custodian of the dataset, usually first name + lastname. The string may contain a list of persons, which should then be separated by semicolons.",
            read_only=False,
            required_by_derived=True,
            required_by_raw=True,
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="owner_email",
            description="Email of owner or of custodian of the dataset. The string may contain a list of emails, which should then be separated by semicolons.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="owner_group",
            description="Defines the group which owns the data, and therefore has unrestricted access to this data. Usually a pgroup like p12151.",
            read_only=False,
            required_by_derived=True,
            required_by_raw=True,
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="packed_size",
            description="Total size of all datablock package files created for this dataset.",
            read_only=True,
            required_by_derived=False,
            required_by_raw=False,
            type=int,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="pid",
            description="Persistent identifier for datasets.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=PID,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="principal_investigator",
            description="Name(s) of the principal investigator(s). The string may contain a list of names, which should then be separated by semicolons.",
            read_only=False,
            required_by_derived=True,
            required_by_raw=True,
            type=str,
            used_by_derived=False,
            used_by_raw=True,
        ),
        Field(
            name="proposal_id",
            description="Identifier for the proposal that the dataset was produced for.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=False,
            used_by_raw=True,
        ),
        Field(
            name="sample_id",
            description="Identifier for the sample that the dataset contains a measurement of.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=False,
            used_by_raw=True,
        ),
        Field(
            name="shared_with",
            description="List of users that the dataset has been shared with.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=List[str],
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="size",
            description="Total size of all files contained in source folder on disk when unpacked.",
            read_only=True,
            required_by_derived=False,
            required_by_raw=False,
            type=int,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="source_folder",
            description="Absolute file path on file server containing the files of this dataset, e.g. /some/path/to/sourcefolder. In case of a single file dataset, e.g. HDF5 data, it contains the path up to, but excluding the filename.",
            read_only=False,
            required_by_derived=True,
            required_by_raw=True,
            type=RemotePath,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="source_folder_host",
            description="DNS host name of file server hosting source_folder, optionally including protocol e.g. [protocol://]fileserver1.example.com",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="techniques",
            description="List of techniques used to produce the data.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=List[Technique],
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="type",
            description="Dataset type. 'Derived' or 'Raw'",
            read_only=False,
            required_by_derived=True,
            required_by_raw=True,
            type=DatasetType,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="updated_at",
            description="Time when the object was last updated in the database.",
            read_only=True,
            required_by_derived=False,
            required_by_raw=False,
            type=datetime,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="updated_by",
            description="Account name who last updated the object in the database.",
            read_only=True,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="used_software",
            description="A list of links to software repositories which uniquely identifies the software used and the version for producing the derived data.",
            read_only=False,
            required_by_derived=True,
            required_by_raw=True,
            type=List[str],
            used_by_derived=True,
            used_by_raw=False,
        ),
        Field(
            name="validation_status",
            description="Defines a level of trust, e.g. a measure of how much data was verified or used by other persons.",
            read_only=False,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="version",
            description="Version of SciCat API used in creation of dataset.",
            read_only=True,
            required_by_derived=False,
            required_by_raw=False,
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
    ]

    def __init__(
        self,
        *,
        type: Union[DatasetType, Literal["derived", "raw"]],
        creation_time: Optional[Union[datetime, str]] = None,
        pid: Optional[Union[str, PID]] = None,
        access_groups: Optional[List[str]] = None,
        classification: Optional[str] = None,
        contact_email: Optional[str] = None,
        creation_location: Optional[str] = None,
        data_format: Optional[str] = None,
        description: Optional[str] = None,
        end_time: Optional[datetime] = None,
        input_datasets: Optional[List[PID]] = None,
        instrument_group: Optional[str] = None,
        instrument_id: Optional[str] = None,
        investigator: Optional[str] = None,
        is_published: Optional[bool] = None,
        job_log_data: Optional[str] = None,
        job_parameters: Optional[dict] = None,
        keywords: Optional[List[str]] = None,
        license: Optional[str] = None,
        lifecycle: Optional[DatasetLifecycle] = None,
        meta: Optional[Dict] = None,
        name: Optional[str] = None,
        orcid_of_owner: Optional[str] = None,
        owner: Optional[str] = None,
        owner_email: Optional[str] = None,
        owner_group: Optional[str] = None,
        principal_investigator: Optional[str] = None,
        proposal_id: Optional[str] = None,
        sample_id: Optional[str] = None,
        shared_with: Optional[List[str]] = None,
        source_folder: Optional[RemotePath] = None,
        source_folder_host: Optional[str] = None,
        techniques: Optional[List[Technique]] = None,
        used_software: Optional[List[str]] = None,
        validation_status: Optional[str] = None,
        checksum_algorithm: Optional[str] = "md5",
        _read_only: Optional[Dict[str, Any]] = None,
        _orig_datablocks: Optional[List[OrigDatablockProxy]] = None,
    ):
        """Construct a dataset with given values.

        All arguments correspond to dataset fields, except `checksum_algorithm` which
        is used as the default algorithm when files are added.

        Arguments starting with an underscore are used internally to initialize
        datasets from SciCat models, and you should generally avoid setting them
        yourself!
        """
        _read_only = _read_only or {}
        self._fields = {
            "creation_time": _parse_datetime(creation_time),
            "history": _apply_default(_read_only.get("history"), None, list),
            "pid": PID.parse(pid) if isinstance(pid, str) else pid,
            "type": DatasetType(type),
            "access_groups": _apply_default(access_groups, None, None),
            "classification": _apply_default(classification, None, None),
            "contact_email": _apply_default(contact_email, None, None),
            "created_at": _apply_default(_read_only.get("created_at"), None, None),
            "created_by": _apply_default(_read_only.get("created_by"), None, None),
            "creation_location": _apply_default(creation_location, None, None),
            "data_format": _apply_default(data_format, None, None),
            "description": _apply_default(description, None, None),
            "end_time": _apply_default(end_time, None, None),
            "input_datasets": _apply_default(input_datasets, None, None),
            "instrument_group": _apply_default(instrument_group, None, None),
            "instrument_id": _apply_default(instrument_id, None, None),
            "investigator": _apply_default(investigator, None, None),
            "is_published": _apply_default(is_published, False, None),
            "job_log_data": _apply_default(job_log_data, None, None),
            "job_parameters": _apply_default(job_parameters, None, None),
            "keywords": _apply_default(keywords, None, None),
            "license": _apply_default(license, None, None),
            "lifecycle": _apply_default(lifecycle, None, None),
            "meta": _apply_default(meta, None, dict),
            "name": _apply_default(name, None, None),
            "orcid_of_owner": _apply_default(orcid_of_owner, None, None),
            "owner": _apply_default(owner, None, None),
            "owner_email": _apply_default(owner_email, None, None),
            "owner_group": _apply_default(owner_group, None, None),
            "principal_investigator": _apply_default(
                principal_investigator, None, None
            ),
            "proposal_id": _apply_default(proposal_id, None, None),
            "sample_id": _apply_default(sample_id, None, None),
            "shared_with": _apply_default(shared_with, None, None),
            "source_folder": _apply_default(source_folder, None, None),
            "source_folder_host": _apply_default(source_folder_host, None, None),
            "techniques": _apply_default(techniques, None, None),
            "updated_at": _apply_default(_read_only.get("updated_at"), None, None),
            "updated_by": _apply_default(_read_only.get("updated_by"), None, None),
            "used_software": _apply_default(used_software, None, None),
            "validation_status": _apply_default(validation_status, None, None),
            "version": _apply_default(_read_only.get("version"), None, None),
        }
        self._orig_datablocks = (
            [] if _orig_datablocks is None else list(_orig_datablocks)
        )
        self._default_checksum_algorithm = self._validate_checksum_algorithm(
            checksum_algorithm
        )

    @staticmethod
    def _validate_checksum_algorithm(algorithm: Optional[str]) -> Optional[str]:
        if algorithm is None:
            return algorithm
        import hashlib

        if algorithm not in hashlib.algorithms_available:
            raise ValueError(f"Checksum algorithm not recognized: {algorithm}")
        return algorithm

    @property
    def type(self) -> DatasetType:
        """Dataset type, Derived or Raw."""
        return self._fields["type"]  # type: ignore[no-any-return]

    @type.setter
    def type(self, val: DatasetType) -> None:
        """Dataset type, Derived or Raw."""
        self._fields["type"] = val

    @property
    def pid(self) -> Optional[PID]:
        """Persistent identifier for datasets."""
        return self._fields["pid"]  # type: ignore[no-any-return]

    @pid.setter
    def pid(self, pid: Optional[Union[PID, str]]) -> None:
        """Persistent identifier for datasets."""
        self._fields["pid"] = None if pid is None else PID.parse(pid)

    @property
    def creation_time(self) -> datetime:
        """Time when the dataset became fully available on disk,
        i.e. all containing files have been written."""
        return self._fields["creation_time"]  # type: ignore[no-any-return]

    @creation_time.setter
    def creation_time(self, value: Union[datetime, str]) -> None:
        """Time when the dataset became fully available on disk,
        i.e. all containing files have been written."""
        if value is None:
            raise TypeError("Cannot set creation_time to None")
        self._fields["creation_time"] = _parse_datetime(value)

    @property
    def access_groups(self) -> Optional[List[str]]:
        """Groups which have read access to the data. The special group 'public' makes data available to all users."""
        return self._fields["access_groups"]  # type: ignore[no-any-return]

    @access_groups.setter
    def access_groups(self, val: Optional[List[str]]) -> None:
        self._fields["access_groups"] = val

    @property
    def classification(self) -> Optional[str]:
        """ACIA information about AUthenticity,COnfidentiality,INtegrity and AVailability requirements of dataset. E.g. AV(ailabilty)=medium could trigger the creation of two tape copies. Format 'AV=medium,CO=low'"""
        return self._fields["classification"]  # type: ignore[no-any-return]

    @classification.setter
    def classification(self, val: Optional[str]) -> None:
        self._fields["classification"] = val

    @property
    def contact_email(self) -> Optional[str]:
        """Email of contact person for this dataset. May contain a list of emails, which should then be separated by semicolons."""
        return self._fields["contact_email"]  # type: ignore[no-any-return]

    @contact_email.setter
    def contact_email(self, val: Optional[str]) -> None:
        self._fields["contact_email"] = val

    @property
    def created_at(self) -> Optional[datetime]:
        """Time when the object was created in the database."""
        return self._fields["created_at"]  # type: ignore[no-any-return]

    @property
    def created_by(self) -> Optional[str]:
        """Account name who created the object in the database."""
        return self._fields["created_by"]  # type: ignore[no-any-return]

    @property
    def creation_location(self) -> Optional[str]:
        """Unique location identifier where data was taken, usually in the form /Site-name/facility-name/instrumentOrBeamline-name"""
        return self._fields["creation_location"]  # type: ignore[no-any-return]

    @creation_location.setter
    def creation_location(self, val: Optional[str]) -> None:
        self._fields["creation_location"] = val

    @property
    def data_format(self) -> Optional[str]:
        """Defines format of subsequent scientific meta data, e.g Nexus Version x.y"""
        return self._fields["data_format"]  # type: ignore[no-any-return]

    @data_format.setter
    def data_format(self, val: Optional[str]) -> None:
        self._fields["data_format"] = val

    @property
    def description(self) -> Optional[str]:
        """Free text explanation of the contents of the dataset."""
        return self._fields["description"]  # type: ignore[no-any-return]

    @description.setter
    def description(self, val: Optional[str]) -> None:
        self._fields["description"] = val

    @property
    def end_time(self) -> Optional[datetime]:
        """Time of end of data taking for this dataset."""
        return self._fields["end_time"]  # type: ignore[no-any-return]

    @end_time.setter
    def end_time(self, val: Optional[datetime]) -> None:
        self._fields["end_time"] = val

    @property
    def input_datasets(self) -> Optional[List[PID]]:
        """Array of input dataset identifiers used in producing the derived dataset. Ideally these are the global identifier to existing datasets inside this or federated data catalogs."""
        return self._fields["input_datasets"]  # type: ignore[no-any-return]

    @input_datasets.setter
    def input_datasets(self, val: Optional[List[PID]]) -> None:
        self._fields["input_datasets"] = val

    @property
    def instrument_group(self) -> Optional[str]:
        """Groups which have read and write access to the data."""
        return self._fields["instrument_group"]  # type: ignore[no-any-return]

    @instrument_group.setter
    def instrument_group(self, val: Optional[str]) -> None:
        self._fields["instrument_group"] = val

    @property
    def instrument_id(self) -> Optional[str]:
        """SciCat ID of the instrument used to measure the data."""
        return self._fields["instrument_id"]  # type: ignore[no-any-return]

    @instrument_id.setter
    def instrument_id(self, val: Optional[str]) -> None:
        self._fields["instrument_id"] = val

    @property
    def investigator(self) -> Optional[str]:
        """Name(s) of the investigator(s). The string may contain a list of names, which should then be separated by semicolons."""
        return self._fields["investigator"]  # type: ignore[no-any-return]

    @investigator.setter
    def investigator(self, val: Optional[str]) -> None:
        self._fields["investigator"] = val

    @property
    def is_published(self) -> Optional[bool]:
        """Indicate whether the dataset is publicly available."""
        return self._fields["is_published"]  # type: ignore[no-any-return]

    @is_published.setter
    def is_published(self, val: Optional[bool]) -> None:
        self._fields["is_published"] = val

    @property
    def job_log_data(self) -> Optional[str]:
        """The output job logfile. Keep the size of this log data well below 15 MB."""
        return self._fields["job_log_data"]  # type: ignore[no-any-return]

    @job_log_data.setter
    def job_log_data(self, val: Optional[str]) -> None:
        self._fields["job_log_data"] = val

    @property
    def job_parameters(self) -> Optional[dict]:
        """Input parameters to the job that produced the derived data."""
        return self._fields["job_parameters"]  # type: ignore[no-any-return]

    @job_parameters.setter
    def job_parameters(self, val: Optional[dict]) -> None:
        self._fields["job_parameters"] = val

    @property
    def keywords(self) -> Optional[List[str]]:
        """Array of tags associated with the meaning or contents of this dataset. Values should ideally come from defined vocabularies, taxonomies, ontologies or knowledge graphs."""
        return self._fields["keywords"]  # type: ignore[no-any-return]

    @keywords.setter
    def keywords(self, val: Optional[List[str]]) -> None:
        self._fields["keywords"] = val

    @property
    def license(self) -> Optional[str]:
        """Name of license under which data can be used."""
        return self._fields["license"]  # type: ignore[no-any-return]

    @license.setter
    def license(self, val: Optional[str]) -> None:
        self._fields["license"] = val

    @property
    def lifecycle(self) -> Optional[DatasetLifecycle]:
        """Parameters for storage and publishing of dataset."""
        return self._fields["lifecycle"]  # type: ignore[no-any-return]

    @lifecycle.setter
    def lifecycle(self, val: Optional[DatasetLifecycle]) -> None:
        self._fields["lifecycle"] = val

    @property
    def meta(self) -> Optional[Dict]:
        """Free form meta data."""
        return self._fields["meta"]  # type: ignore[no-any-return]

    @meta.setter
    def meta(self, val: Optional[Dict]) -> None:
        self._fields["meta"] = val

    @property
    def name(self) -> Optional[str]:
        """A name for the dataset."""
        return self._fields["name"]  # type: ignore[no-any-return]

    @name.setter
    def name(self, val: Optional[str]) -> None:
        self._fields["name"] = val

    @property
    def orcid_of_owner(self) -> Optional[str]:
        """ORCID of owner/custodian. The string may contain a list of ORCID, which should then be separated by semicolons. ORCIDs must include the prefix https://orcid.org/"""
        return self._fields["orcid_of_owner"]  # type: ignore[no-any-return]

    @orcid_of_owner.setter
    def orcid_of_owner(self, val: Optional[str]) -> None:
        self._fields["orcid_of_owner"] = val

    @property
    def owner(self) -> Optional[str]:
        """Owner or custodian of the dataset, usually first name + lastname. The string may contain a list of persons, which should then be separated by semicolons."""
        return self._fields["owner"]  # type: ignore[no-any-return]

    @owner.setter
    def owner(self, val: Optional[str]) -> None:
        self._fields["owner"] = val

    @property
    def owner_email(self) -> Optional[str]:
        """Email of owner or of custodian of the dataset. The string may contain a list of emails, which should then be separated by semicolons."""
        return self._fields["owner_email"]  # type: ignore[no-any-return]

    @owner_email.setter
    def owner_email(self, val: Optional[str]) -> None:
        self._fields["owner_email"] = val

    @property
    def owner_group(self) -> Optional[str]:
        """Defines the group which owns the data, and therefore has unrestricted access to this data. Usually a pgroup like p12151."""
        return self._fields["owner_group"]  # type: ignore[no-any-return]

    @owner_group.setter
    def owner_group(self, val: Optional[str]) -> None:
        self._fields["owner_group"] = val

    @property
    def principal_investigator(self) -> Optional[str]:
        """Name(s) of the principal investigator(s). The string may contain a list of names, which should then be separated by semicolons."""
        return self._fields["principal_investigator"]  # type: ignore[no-any-return]

    @principal_investigator.setter
    def principal_investigator(self, val: Optional[str]) -> None:
        self._fields["principal_investigator"] = val

    @property
    def proposal_id(self) -> Optional[str]:
        """Identifier for the proposal that the dataset was produced for."""
        return self._fields["proposal_id"]  # type: ignore[no-any-return]

    @proposal_id.setter
    def proposal_id(self, val: Optional[str]) -> None:
        self._fields["proposal_id"] = val

    @property
    def sample_id(self) -> Optional[str]:
        """Identifier for the sample that the dataset contains a measurement of."""
        return self._fields["sample_id"]  # type: ignore[no-any-return]

    @sample_id.setter
    def sample_id(self, val: Optional[str]) -> None:
        self._fields["sample_id"] = val

    @property
    def shared_with(self) -> Optional[List[str]]:
        """List of users that the dataset has been shared with."""
        return self._fields["shared_with"]  # type: ignore[no-any-return]

    @shared_with.setter
    def shared_with(self, val: Optional[List[str]]) -> None:
        self._fields["shared_with"] = val

    @property
    def source_folder(self) -> Optional[RemotePath]:
        """Absolute file path on file server containing the files of this dataset, e.g. /some/path/to/sourcefolder. In case of a single file dataset, e.g. HDF5 data, it contains the path up to, but excluding the filename."""
        return self._fields["source_folder"]  # type: ignore[no-any-return]

    @source_folder.setter
    def source_folder(self, val: Optional[RemotePath]) -> None:
        self._fields["source_folder"] = val

    @property
    def source_folder_host(self) -> Optional[str]:
        """DNS host name of file server hosting source_folder, optionally including protocol e.g. [protocol://]fileserver1.example.com"""
        return self._fields["source_folder_host"]  # type: ignore[no-any-return]

    @source_folder_host.setter
    def source_folder_host(self, val: Optional[str]) -> None:
        self._fields["source_folder_host"] = val

    @property
    def techniques(self) -> Optional[List[Technique]]:
        """List of techniques used to produce the data."""
        return self._fields["techniques"]  # type: ignore[no-any-return]

    @techniques.setter
    def techniques(self, val: Optional[List[Technique]]) -> None:
        self._fields["techniques"] = val

    @property
    def updated_at(self) -> Optional[datetime]:
        """Time when the object was last updated in the database."""
        return self._fields["updated_at"]  # type: ignore[no-any-return]

    @property
    def updated_by(self) -> Optional[str]:
        """Account name who last updated the object in the database."""
        return self._fields["updated_by"]  # type: ignore[no-any-return]

    @property
    def used_software(self) -> Optional[List[str]]:
        """A list of links to software repositories which uniquely identifies the software used and the version for producing the derived data."""
        return self._fields["used_software"]  # type: ignore[no-any-return]

    @used_software.setter
    def used_software(self, val: Optional[List[str]]) -> None:
        self._fields["used_software"] = val

    @property
    def validation_status(self) -> Optional[str]:
        """Defines a level of trust, e.g. a measure of how much data was verified or used by other persons."""
        return self._fields["validation_status"]  # type: ignore[no-any-return]

    @validation_status.setter
    def validation_status(self, val: Optional[str]) -> None:
        self._fields["validation_status"] = val

    @property
    def version(self) -> Optional[str]:
        """Version of SciCat API used in creation of dataset."""
        return self._fields["version"]  # type: ignore[no-any-return]

    def _make_derived_model(self) -> DerivedDataset:
        extra_fields = {
            name: None
            for name in (
                "creation_location",
                "data_format",
                "end_time",
                "instrument_id",
                "principal_investigator",
                "proposal_id",
                "sample_id",
            )
            if getattr(self, name, None) is not None
        }
        return DerivedDataset(
            accessGroups=self.access_groups,
            classification=self.classification,
            contactEmail=self.contact_email,
            createdAt=self.created_at,
            createdBy=self.created_by,
            creationTime=self.creation_time,
            description=self.description,
            history=self.history,
            inputDatasets=self.input_datasets,
            instrumentGroup=self.instrument_group,
            investigator=self.investigator,
            isPublished=self.is_published,
            jobLogData=self.job_log_data,
            jobParameters=self.job_parameters,
            keywords=self.keywords,
            license=self.license,
            datasetlifecycle=self.lifecycle,
            scientificMetadata=self.meta,
            datasetName=self.name,
            numberOfFiles=self.number_of_files,
            numberOfFilesArchived=self.number_of_files_archived,
            orcidOfOwner=self.orcid_of_owner,
            owner=self.owner,
            ownerEmail=self.owner_email,
            ownerGroup=self.owner_group,
            packedSize=self.packed_size,
            pid=self.pid,
            sharedWith=self.shared_with,
            size=self.size,
            sourceFolder=self.source_folder,
            sourceFolderHost=self.source_folder_host,
            techniques=self.techniques,
            type=self.type,
            updatedAt=self.updated_at,
            updatedBy=self.updated_by,
            usedSoftware=self.used_software,
            validationStatus=self.validation_status,
            version=self.version,
            **extra_fields,
        )

    def _make_raw_model(self) -> RawDataset:
        extra_fields = {
            name: None
            for name in (
                "input_datasets",
                "investigator",
                "job_log_data",
                "job_parameters",
                "used_software",
            )
            if getattr(self, name, None) is not None
        }
        return RawDataset(
            accessGroups=self.access_groups,
            classification=self.classification,
            contactEmail=self.contact_email,
            createdAt=self.created_at,
            createdBy=self.created_by,
            creationLocation=self.creation_location,
            creationTime=self.creation_time,
            dataFormat=self.data_format,
            description=self.description,
            endTime=self.end_time,
            history=self.history,
            instrumentGroup=self.instrument_group,
            instrumentId=self.instrument_id,
            isPublished=self.is_published,
            keywords=self.keywords,
            license=self.license,
            datasetlifecycle=self.lifecycle,
            scientificMetadata=self.meta,
            datasetName=self.name,
            numberOfFiles=self.number_of_files,
            numberOfFilesArchived=self.number_of_files_archived,
            orcidOfOwner=self.orcid_of_owner,
            owner=self.owner,
            ownerEmail=self.owner_email,
            ownerGroup=self.owner_group,
            packedSize=self.packed_size,
            pid=self.pid,
            principalInvestigator=self.principal_investigator,
            proposalId=self.proposal_id,
            sampleId=self.sample_id,
            sharedWith=self.shared_with,
            size=self.size,
            sourceFolder=self.source_folder,
            sourceFolderHost=self.source_folder_host,
            techniques=self.techniques,
            type=self.type,
            updatedAt=self.updated_at,
            updatedBy=self.updated_by,
            validationStatus=self.validation_status,
            version=self.version,
            **extra_fields,
        )


def fields_from_model(model: Union[DerivedDataset, RawDataset]) -> dict:
    return (
        _fields_from_derived_model(model)
        if isinstance(model, DerivedDataset)
        else _fields_from_raw_model(model)
    )


def _fields_from_derived_model(model) -> dict:
    return dict(
        _read_only=dict(
            created_at=model.createdAt,
            created_by=model.createdBy,
            updated_at=model.updatedAt,
            updated_by=model.updatedBy,
            version=model.version,
        ),
        access_groups=model.accessGroups,
        classification=model.classification,
        contact_email=model.contactEmail,
        description=model.description,
        input_datasets=model.inputDatasets,
        instrument_group=model.instrumentGroup,
        investigator=model.investigator,
        is_published=model.isPublished,
        job_log_data=model.jobLogData,
        job_parameters=model.jobParameters,
        keywords=model.keywords,
        license=model.license,
        lifecycle=model.datasetlifecycle,
        meta=model.scientificMetadata,
        name=model.datasetName,
        orcid_of_owner=model.orcidOfOwner,
        owner=model.owner,
        owner_email=model.ownerEmail,
        owner_group=model.ownerGroup,
        shared_with=model.sharedWith,
        source_folder=model.sourceFolder,
        source_folder_host=model.sourceFolderHost,
        techniques=model.techniques,
        used_software=model.usedSoftware,
        validation_status=model.validationStatus,
    )


def _fields_from_raw_model(model) -> dict:
    return dict(
        _read_only=dict(
            created_at=model.createdAt,
            created_by=model.createdBy,
            updated_at=model.updatedAt,
            updated_by=model.updatedBy,
            version=model.version,
        ),
        access_groups=model.accessGroups,
        classification=model.classification,
        contact_email=model.contactEmail,
        creation_location=model.creationLocation,
        data_format=model.dataFormat,
        description=model.description,
        end_time=model.endTime,
        instrument_group=model.instrumentGroup,
        instrument_id=model.instrumentId,
        is_published=model.isPublished,
        keywords=model.keywords,
        license=model.license,
        lifecycle=model.datasetlifecycle,
        meta=model.scientificMetadata,
        name=model.datasetName,
        orcid_of_owner=model.orcidOfOwner,
        owner=model.owner,
        owner_email=model.ownerEmail,
        owner_group=model.ownerGroup,
        principal_investigator=model.principalInvestigator,
        proposal_id=model.proposalId,
        sample_id=model.sampleId,
        shared_with=model.sharedWith,
        source_folder=model.sourceFolder,
        source_folder_host=model.sourceFolderHost,
        techniques=model.techniques,
        validation_status=model.validationStatus,
    )



##########################################
# This file was automatically generated. #
# Do not modify it directly!             #
##########################################
# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
# flake8: noqa

"""Pydantic models to encode data for communication with SciCat."""

import enum
import os
from datetime import datetime
from typing import Any, Dict, List, Optional, Type, TypeVar

import pydantic

from ._internal.orcid import is_valid_orcid
from .filesystem import RemotePath
from .logging import get_logger
from .pid import PID


# This can be replaced by StrEnum in Python 3.11+
class DatasetType(str, enum.Enum):
    """Type of Dataset"""

    RAW = "raw"
    DERIVED = "derived"


class BaseModel(pydantic.BaseModel):
    class Config:
        extra = pydantic.Extra.forbid
        json_encoders = {
            PID: lambda v: str(v),
            RemotePath: lambda v: os.fspath(v),  # type: ignore[no-any-return]
        }


class DatasetLifecycle(BaseModel):
    archivable: Optional[bool]
    archiveRetentionTime: Optional[datetime]
    archiveReturnMessage: Optional[str]
    archiveStatusMessage: Optional[str]
    dateOfDiskPurging: Optional[datetime]
    dateOfPublishing: Optional[datetime]
    exportedTo: Optional[str]
    isOnCentralDisk: Optional[bool]
    publishable: Optional[bool]
    publishedOn: Optional[datetime]
    retrievable: Optional[bool]
    retrieveIntegrityCheck: Optional[bool]
    retrieveReturnMessage: Optional[str]
    retrieveStatusMessage: Optional[str]


class MongoQueryable(BaseModel):
    createdAt: Optional[datetime]
    createdBy: Optional[str]
    updatedAt: Optional[datetime]
    updatedBy: Optional[str]


class Technique(BaseModel):
    name: str
    pid: str


class DataFile(MongoQueryable):
    path: str
    size: int
    chk: Optional[str]
    gid: Optional[str]
    perm: Optional[str]
    time: Optional[datetime]
    uid: Optional[str]

    @pydantic.validator("size")
    def _validate_size(cls, value: Any) -> Any:
        return _validate_size(value)


class Ownable(MongoQueryable):
    ownerGroup: str
    accessGroups: Optional[List[str]]
    instrumentGroup: Optional[str]


class Datablock(Ownable):
    archiveId: str
    dataFileList: List[DataFile]
    size: int
    version: str
    chkAlg: Optional[str]
    datasetId: Optional[str]
    packedSize: Optional[int]
    id: Optional[str]

    @pydantic.validator("size", "packedSize")
    def _validate_size(cls, value: Any) -> Any:
        return _validate_size(value)


class DerivedDataset(Ownable):
    contactEmail: str
    creationTime: datetime
    inputDatasets: List[PID]
    investigator: str
    owner: str
    sourceFolder: RemotePath
    type: DatasetType
    usedSoftware: List[str]
    classification: Optional[str]
    description: Optional[str]
    history: Optional[List[dict]]
    isPublished: Optional[bool]
    jobLogData: Optional[str]
    jobParameters: Optional[dict]
    keywords: Optional[List[str]]
    license: Optional[str]
    datasetlifecycle: Optional[DatasetLifecycle]
    scientificMetadata: Optional[Dict]
    datasetName: Optional[str]
    numberOfFiles: Optional[int]
    numberOfFilesArchived: Optional[int]
    orcidOfOwner: Optional[str]
    ownerEmail: Optional[str]
    packedSize: Optional[int]
    pid: Optional[PID]
    sharedWith: Optional[List[str]]
    size: Optional[int]
    sourceFolderHost: Optional[str]
    techniques: Optional[List[Technique]]
    validationStatus: Optional[str]
    version: Optional[str]

    @pydantic.validator("contactEmail", "ownerEmail")
    def _validate_emails(cls, value: Any) -> Any:
        return _validate_emails(value)

    @pydantic.validator("numberOfFiles", "numberOfFilesArchived", "packedSize", "size")
    def _validate_size(cls, value: Any) -> Any:
        return _validate_size(value)

    @pydantic.validator("orcidOfOwner")
    def _validate_orcid(cls, value: Any) -> Any:
        return _validate_orcid(value)


class OrigDatablock(Ownable):
    dataFileList: List[DataFile]
    size: int
    datasetId: Optional[PID]
    id: Optional[PID]

    @pydantic.validator("size")
    def _validate_size(cls, value: Any) -> Any:
        return _validate_size(value)


class RawDataset(Ownable):
    contactEmail: str
    creationTime: datetime
    owner: str
    principalInvestigator: str
    sourceFolder: RemotePath
    type: DatasetType
    classification: Optional[str]
    creationLocation: Optional[str]
    dataFormat: Optional[str]
    description: Optional[str]
    endTime: Optional[datetime]
    history: Optional[List[dict]]
    instrumentId: Optional[str]
    isPublished: Optional[bool]
    keywords: Optional[List[str]]
    license: Optional[str]
    datasetlifecycle: Optional[DatasetLifecycle]
    scientificMetadata: Optional[Dict]
    datasetName: Optional[str]
    numberOfFiles: Optional[int]
    numberOfFilesArchived: Optional[int]
    orcidOfOwner: Optional[str]
    ownerEmail: Optional[str]
    packedSize: Optional[int]
    pid: Optional[PID]
    proposalId: Optional[str]
    sampleId: Optional[str]
    sharedWith: Optional[List[str]]
    size: Optional[int]
    sourceFolderHost: Optional[str]
    techniques: Optional[List[Technique]]
    validationStatus: Optional[str]
    version: Optional[str]

    @pydantic.validator("contactEmail", "ownerEmail")
    def _validate_emails(cls, value: Any) -> Any:
        return _validate_emails(value)

    @pydantic.validator("numberOfFiles", "numberOfFilesArchived", "packedSize", "size")
    def _validate_size(cls, value: Any) -> Any:
        return _validate_size(value)

    @pydantic.validator("orcidOfOwner")
    def _validate_orcid(cls, value: Any) -> Any:
        return _validate_orcid(value)


def _validate_emails(value: Optional[str]) -> Optional[str]:
    if value is None:
        return value
    return ";".join(pydantic.EmailStr.validate(item) for item in value.split(";"))


def _validate_size(value: Optional[int]) -> Optional[int]:
    if value is None:
        return value
    if value < 0:
        raise ValueError("Must be > 0")
    return value


def _validate_orcid(value: Optional[str]) -> Optional[str]:
    if value is None:
        return value
    try:
        if is_valid_orcid(value):
            return value
    except (RuntimeError, ValueError, TypeError):
        pass
    raise ValueError(
        "value is not a valid ORCID, "
        "note that ORCIDs must be prefixed with 'https://orcid.org'."
    )


ModelType = TypeVar("ModelType", bound=pydantic.BaseModel)


def construct(
    model: Type[ModelType], *, _strict_validation: bool = True, **fields: Any
) -> ModelType:
    """Instantiates a model.

    Warning
    -------
    If the model is created without validation, no fields will be converted
    to their proper type but will simply be whatever arguments are passed.

    A warning will be emitted in this case.

    Parameters
    ----------
    model:
        Class of the model to create.
    _strict_validation:
        If ``True``, the model must pass validation.
        If ``False``, a model is still returned if validation fails.
    fields:
        Field values to pass to the model initializer.

    Returns
    -------
    :
        An initialized model.
    """
    try:
        return model(**fields)
    except pydantic.ValidationError as e:
        if _strict_validation:
            raise
        get_logger().warning(
            "Validation of metadata failed: %s\n"
            "The returned object may be incomplete or broken. "
            "In particular, some fields may not have the correct type",
            str(e),
        )
        return model.construct(**fields)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
import subprocess
from contextlib import contextmanager
from pathlib import Path


def docker_compose_up(config_file: Path, *services: str):
    subprocess.check_call(
        [
            "docker",
            "compose",
            "--file",
            str(config_file),
            "up",
            "--detach",
            "--force-recreate",
            *services,
        ]
    )


def docker_compose_down(config_file: Path):
    subprocess.check_call(
        ["docker", "compose", "--file", str(config_file), "down", "--volumes"]
    )


@contextmanager
def docker_compose(config_file: Path, *services: str):
    docker_compose_up(config_file, *services)
    try:
        yield
    finally:
        docker_compose_down(config_file)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
import uuid

import pytest
from dateutil.parser import parse as parse_date

from scitacean import PID, Client, DatasetType
from scitacean.model import DerivedDataset, RawDataset, Technique

from .common.backend import skip_if_not_backend


@pytest.fixture
def client(
    request,
    scicat_access,
    scicat_backend,
):
    skip_if_not_backend(request)
    return Client.from_credentials(
        url=scicat_access.url, **scicat_access.functional_credentials
    )


def test_derived_dataset_default_values(client):
    dset = DerivedDataset(
        accessGroups=["access1"],
        contactEmail="contact@email.com",
        creationTime=parse_date("2000-01-01T01:01:01.000Z"),
        inputDatasets=[PID(prefix="PID.SAMPLE.PREFIX", pid="abcd")],
        investigator="inv@esti.gator",
        owner="owner",
        ownerGroup="ownerGroup",
        sourceFolder="/source/folder",
        usedSoftware=["software1"],
        type=DatasetType.DERIVED,
    )
    pid = client.scicat.create_dataset_model(dset).pid
    finalized = client.scicat.get_dataset_model(pid)

    # Inputs
    assert finalized.accessGroups == ["access1"]
    assert finalized.contactEmail == "contact@email.com"
    assert finalized.creationTime == parse_date("2000-01-01T01:01:01.000Z")
    assert finalized.inputDatasets == [PID(prefix="PID.SAMPLE.PREFIX", pid="abcd")]
    assert finalized.investigator == "inv@esti.gator"
    assert finalized.owner == "owner"
    assert finalized.ownerGroup == "ownerGroup"
    assert finalized.sourceFolder == "/source/folder"
    assert finalized.usedSoftware == ["software1"]

    # Default values
    assert finalized.createdAt  # some non-empty str
    assert finalized.createdBy  # some non-empty str
    assert finalized.classification  # some non-empty str
    assert finalized.datasetName  # some non-empty str
    assert finalized.history == []
    assert finalized.isPublished is None
    assert finalized.pid  # some non-empty str
    assert finalized.techniques == []
    assert finalized.updatedAt  # some non-empty str
    assert finalized.version  # some non-empty str

    # Left empty
    assert finalized.description is None is None
    assert finalized.jobParameters is None
    assert finalized.jobLogData is None
    assert finalized.keywords is None
    assert finalized.license is None
    assert finalized.numberOfFiles is None
    assert finalized.numberOfFilesArchived is None
    assert finalized.orcidOfOwner is None
    assert finalized.ownerEmail is None
    assert finalized.packedSize is None
    assert finalized.scientificMetadata is None
    assert finalized.sharedWith is None
    assert finalized.size is None
    assert finalized.sourceFolderHost is None
    assert finalized.validationStatus is None


def test_raw_dataset_default_values(client):
    dset = RawDataset(
        accessGroups=["access1"],
        contactEmail="contact@email.com",
        creationTime=parse_date("2000-01-01T01:01:01.000Z"),
        owner="owner",
        ownerGroup="ownerGroup",
        principalInvestigator="inv@esti.gator",
        sourceFolder="/source/folder",
        type=DatasetType.RAW,
    )
    pid = client.scicat.create_dataset_model(dset).pid
    finalized = client.scicat.get_dataset_model(pid)

    # Inputs
    assert finalized.accessGroups == ["access1"]
    assert finalized.contactEmail == "contact@email.com"
    assert finalized.creationTime == parse_date("2000-01-01T01:01:01.000Z")
    assert finalized.owner == "owner"
    assert finalized.ownerGroup == "ownerGroup"
    assert finalized.principalInvestigator == "inv@esti.gator"
    assert finalized.sourceFolder == "/source/folder"

    # Default values
    assert finalized.createdAt  # some non-empty str
    assert finalized.createdBy  # some non-empty str
    assert finalized.classification  # some non-empty str
    assert finalized.datasetName  # some non-empty str
    assert finalized.history == []
    assert finalized.isPublished is None
    assert finalized.pid  # some non-empty str
    assert finalized.techniques == []
    assert finalized.updatedAt  # some non-empty str
    assert finalized.version  # some non-empty str

    # Left empty
    assert finalized.creationLocation is None
    assert finalized.dataFormat is None
    assert finalized.description is None
    assert finalized.endTime is None
    assert finalized.instrumentId is None
    assert finalized.keywords is None
    assert finalized.license is None
    assert finalized.numberOfFiles is None
    assert finalized.numberOfFilesArchived is None
    assert finalized.orcidOfOwner is None
    assert finalized.ownerEmail is None
    assert finalized.packedSize is None
    assert finalized.proposalId is None
    assert finalized.sampleId is None
    assert finalized.scientificMetadata is None
    assert finalized.sharedWith is None
    assert finalized.size is None
    assert finalized.sourceFolderHost is None
    assert finalized.validationStatus is None


def test_derived_dataset_overwritten_values(client):
    dset = DerivedDataset(
        accessGroups=["access1"],
        classification="classes",
        contactEmail="contact@email.com",
        createdAt=parse_date("2001-01-01T01:01:01.000Z"),
        createdBy="creator",
        creationTime=parse_date("2000-01-01T01:01:01.000Z"),
        datasetName="datasetName",
        description="description",
        history=[
            DerivedDataset(
                accessGroups=["access1"],
                contactEmail="contact@email.com",
                creationTime=parse_date("2000-01-01T01:01:01.000Z"),
                inputDatasets=[PID(prefix="PID.SAMPLE.PREFIX", pid="abcd")],
                investigator="inv@esti.gator",
                owner="owner",
                ownerGroup="ownerGroup",
                sourceFolder="/source/folder",
                usedSoftware=["software1"],
                type=DatasetType.DERIVED,
            ).dict()
        ],
        inputDatasets=[PID(prefix="PID.SAMPLE.PREFIX", pid="abcd")],
        investigator="inv@esti.gator",
        isPublished=True,
        jobParameters={"tiredness": "great"},
        jobLogData="all-good",
        keywords=["keyword1"],
        license="the-license",
        numberOfFiles=123,
        numberOfFilesArchived=42,
        orcidOfOwner="https://orcid.org/0000-0002-3761-3201",
        owner="owner",
        ownerEmail="owner@email.eu",
        ownerGroup="ownerGroup",
        packedSize=551,
        pid=PID(pid=f"derived-id-{uuid.uuid4()}"),
        scientificMetadata={"sci1": ["data1"]},
        sharedWith=["shared1"],
        size=6123,
        sourceFolder="/source/folder",
        sourceFolderHost="/host/source/folder",
        techniques=[Technique(name="technique1", pid="technique1-id")],
        type=DatasetType.DERIVED,
        updatedAt=parse_date("2001-01-01T01:01:01.000Z"),
        usedSoftware=["software1"],
        validationStatus="not-validated",
        version="9999.00.888",
    )
    pid = client.scicat.create_dataset_model(dset).pid
    finalized = client.scicat.get_dataset_model(pid)

    preserved_inputs = [
        "accessGroups",
        "classification",
        "contactEmail",
        "creationTime",
        "datasetName",
        "description",
        "inputDatasets",
        "investigator",
        "isPublished",
        "jobParameters",
        "jobLogData",
        "keywords",
        "license",
        "numberOfFiles",
        "numberOfFilesArchived",
        "orcidOfOwner",
        "owner",
        "ownerEmail",
        "ownerGroup",
        "packedSize",
        "scientificMetadata",
        "sharedWith",
        "size",
        "sourceFolder",
        "sourceFolderHost",
        "techniques",
        "usedSoftware",
        "validationStatus",
    ]

    overwritten_inputs = ["createdAt", "createdBy", "updatedAt", "pid", "version"]

    for key in preserved_inputs:
        assert getattr(finalized, key) == getattr(dset, key)
    for key in overwritten_inputs:
        assert getattr(finalized, key) != getattr(dset, key)
    assert finalized.history == []  # This is a new dataset, so the history is empty.


def test_raw_dataset_overwritten_values(client):
    dset = RawDataset(
        accessGroups=["access1"],
        classification="classes",
        contactEmail="contact@email.com",
        createdAt=parse_date("2001-01-01T01:01:01.000Z"),
        createdBy="creator",
        creationLocation="the-realverse",
        creationTime=parse_date("2000-01-01T01:01:01.000Z"),
        dataFormat="data-format",
        datasetName="datasetName",
        description="description",
        endTime=parse_date("2010-01-01T01:01:01.000Z"),
        history=[
            RawDataset(
                accessGroups=["access1"],
                contactEmail="contact@email.com",
                creationTime=parse_date("2000-01-01T01:01:01.000Z"),
                owner="owner",
                ownerGroup="ownerGroup",
                principalInvestigator="inv@esti.gator",
                sourceFolder="/source/folder",
                type=DatasetType.RAW,
            ).dict()
        ],
        instrumentId="instrument-id",
        isPublished=True,
        keywords=["keyword1"],
        license="the-license",
        numberOfFiles=123,
        numberOfFilesArchived=42,
        orcidOfOwner="https://orcid.org/0000-0002-3761-3201",
        owner="owner",
        ownerEmail="owner@email.eu",
        ownerGroup="ownerGroup",
        packedSize=551,
        pid=PID(pid=f"derived-id-{uuid.uuid4()}"),
        principalInvestigator="inv@esti.gator",
        proposalId="proposal-id",
        sampleId="sample-id",
        scientificMetadata={"sci1": ["data1"]},
        sharedWith=["shared1"],
        size=6123,
        sourceFolder="/source/folder",
        sourceFolderHost="/host/source/folder",
        techniques=[Technique(name="technique1", pid="technique1-id")],
        type=DatasetType.RAW,
        updatedAt=parse_date("2001-01-01T01:01:01.000Z"),
        validationStatus="not-validated",
        version="9999.00.888",
    )
    pid = client.scicat.create_dataset_model(dset).pid
    finalized = client.scicat.get_dataset_model(pid)

    preserved_inputs = [
        "accessGroups",
        "classification",
        "contactEmail",
        "creationLocation",
        "creationTime",
        "dataFormat",
        "datasetName",
        "description",
        "endTime",
        "instrumentId",
        "isPublished",
        "keywords",
        "license",
        "numberOfFiles",
        "numberOfFilesArchived",
        "orcidOfOwner",
        "owner",
        "ownerEmail",
        "ownerGroup",
        "packedSize",
        "principalInvestigator",
        "proposalId",
        "sampleId",
        "scientificMetadata",
        "sharedWith",
        "size",
        "sourceFolder",
        "sourceFolderHost",
        "techniques",
        "validationStatus",
    ]

    overwritten_inputs = ["createdAt", "createdBy", "updatedAt", "pid", "version"]

    for key in preserved_inputs:
        assert getattr(finalized, key) == getattr(dset, key)
    for key in overwritten_inputs:
        assert getattr(finalized, key) != getattr(dset, key)
    assert finalized.history == []  # This is a new dataset, so the history is empty.



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Templates to generate models and other classes."""

from pathlib import Path
from string import Template

TEMPLATE_DIR = Path(__file__).resolve().parent

BANNER = """##########################################
# This file was automatically generated. #
# Do not modify it directly!             #
##########################################
"""


def load_template(name: str) -> Template:
    with open(TEMPLATE_DIR / f"{name}.py.template", "r") as f:
        return Template(f.read())



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

# These tests use Dataset instead of DatasetFields in order to test the
# public interface and make sure that Dataset does not break any behavior.

from datetime import datetime, timedelta, timezone

import dateutil.parser
import pydantic
import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from scitacean import PID, Dataset, DatasetType
from scitacean.filesystem import RemotePath
from scitacean.model import DataFile, DerivedDataset, OrigDatablock, RawDataset


def test_init_dataset_with_only_type():
    dset = Dataset(type="raw")
    assert dset.type == DatasetType.RAW


@pytest.mark.parametrize(
    "typ", ("raw", "derived", DatasetType.RAW, DatasetType.DERIVED)
)
def test_init_dataset_accepted_types(typ):
    dset = Dataset(type=typ)
    assert dset.type == typ


def test_init_dataset_raises_for_bad_type():
    with pytest.raises(ValueError):
        Dataset(type="bad-type")  # type: ignore[arg-type]


def test_init_dataset_needs_type():
    with pytest.raises(TypeError):
        Dataset()  # type: ignore[call-arg]


def test_init_dataset_sets_creation_time():
    expected = datetime.now(tz=timezone.utc)
    dset = Dataset(type="raw")
    assert abs(dset.creation_time - expected) < timedelta(seconds=30)


def test_init_dataset_can_set_creation_time():
    dt = dateutil.parser.parse("2022-01-10T11:14:52.623Z")
    dset = Dataset(type="derived", creation_time=dt)
    assert dset.creation_time == dt

    dt = dateutil.parser.parse("2022-01-10T11:14:52+02:00")
    dset = Dataset(type="derived", creation_time=dt)
    assert dset.creation_time == dt

    dt = "1994-03-21T22:51:33-01:00"
    dset = Dataset(type="derived", creation_time=dt)
    assert dset.creation_time == dateutil.parser.parse(dt)

    dset = Dataset(type="derived", creation_time="now")
    assert abs(dset.creation_time - datetime.now(tz=timezone.utc)) < timedelta(
        seconds=30
    )


def test_init_dataset_default_values():
    dset = Dataset(type="derived")
    assert dset.history == []
    assert dset.is_published is False
    assert dset.meta == {}


@pytest.mark.parametrize("field", Dataset.fields(read_only=True), ids=lambda f: f.name)
def test_cannot_set_read_only_fields(field):
    dset = Dataset(type="raw")
    with pytest.raises(AttributeError):
        setattr(dset, field.name, None)


@pytest.mark.parametrize(
    "field",
    filter(lambda f: f.name != "type", Dataset.fields(read_only=False)),
    ids=lambda f: f.name,
)
@given(st.data())
@settings(max_examples=10)
def test_can_init_writable_fields(field, data):
    value = data.draw(st.from_type(field.type))
    dset = Dataset(type="raw", **{field.name: value})
    assert getattr(dset, field.name) == value


@pytest.mark.parametrize("field", Dataset.fields(read_only=False), ids=lambda f: f.name)
@given(st.data())
@settings(max_examples=10)
def test_can_set_writable_fields(field, data):
    value = data.draw(st.from_type(field.type))
    dset = Dataset(type="raw")
    setattr(dset, field.name, value)
    assert getattr(dset, field.name) == value


@pytest.mark.parametrize(
    "field",
    filter(lambda f: not f.read_only and f.name != "creation_time", Dataset.fields()),
    ids=lambda f: f.name,
)
def test_can_set_writable_fields_to_none(field):
    dset = Dataset(type="raw")
    setattr(dset, field.name, None)
    assert getattr(dset, field.name) is None


def test_cannot_set_creation_time_to_none():
    dset = Dataset(type="raw")
    with pytest.raises(TypeError):
        dset.creation_time = None


def test_init_from_models_sets_metadata():
    dset = Dataset.from_models(
        dataset_model=RawDataset(
            contactEmail="p.stibbons@uu.am",
            creationTime=dateutil.parser.parse("2022-01-10T11:14:52+02:00"),
            principalInvestigator="my principal investigator",
            owner="PonderStibbons",
            sourceFolder=RemotePath("/hex/source91"),
            type=DatasetType.RAW,
            ownerGroup="faculty",
            createdBy="pstibbons",
            createdAt=dateutil.parser.parse("2022-01-10T12:41:22+02:00"),
        ),
        orig_datablock_models=[
            OrigDatablock(
                dataFileList=[],
                size=0,
                ownerGroup="faculty",
            )
        ],
    )

    assert dset.contact_email == "p.stibbons@uu.am"
    assert dset.creation_time == dateutil.parser.parse("2022-01-10T11:14:52+02:00")
    assert dset.principal_investigator == "my principal investigator"
    assert dset.owner == "PonderStibbons"
    assert dset.source_folder == "/hex/source91"
    assert dset.type == DatasetType.RAW
    assert dset.owner_group == "faculty"
    assert dset.created_by == "pstibbons"
    assert dset.created_at == dateutil.parser.parse("2022-01-10T12:41:22+02:00")

    assert dset.is_published is False
    assert dset.owner_email is None

    assert len(list(dset.files)) == 0
    assert dset.number_of_files == 0
    assert dset.number_of_files_archived == 0
    assert dset.packed_size == 0
    assert dset.size == 0


def test_init_from_models_sets_files():
    dset = Dataset.from_models(
        dataset_model=RawDataset(
            contactEmail="p.stibbons@uu.am",
            creationTime=dateutil.parser.parse("2022-01-10T11:14:52-01:00"),
            principalInvestigator="librarian@uu.am",
            owner="PonderStibbons",
            sourceFolder=RemotePath("/hex/source91"),
            type=DatasetType.RAW,
            ownerGroup="faculty",
        ),
        orig_datablock_models=[
            OrigDatablock(
                dataFileList=[
                    DataFile(
                        path="file1.dat",
                        size=6123,
                        time=dateutil.parser.parse("2022-01-09T18:32:01-01:00"),
                    ),
                    DataFile(
                        path="sub/file2.png",
                        size=551,
                        time=dateutil.parser.parse("2022-01-09T18:32:42-01:00"),
                    ),
                ],
                size=6123 + 551,
                ownerGroup="faculty",
            )
        ],
    )

    assert len(list(dset.files)) == 2
    assert dset.number_of_files == 2
    assert dset.number_of_files_archived == 0
    assert dset.packed_size == 0
    assert dset.size == 6123 + 551

    f0 = [f for f in dset.files if f.remote_path.suffix == ".dat"][0]
    assert f0.remote_access_path(dset.source_folder) == "/hex/source91/file1.dat"
    assert f0.local_path is None
    assert f0.size == 6123
    assert f0.make_model().path == "file1.dat"

    f1 = [f for f in dset.files if f.remote_path.suffix == ".png"][0]
    assert f1.remote_access_path(dset.source_folder) == "/hex/source91/sub/file2.png"
    assert f1.local_path is None
    assert f1.size == 551
    assert f1.make_model().path == "sub/file2.png"


def test_init_from_models_sets_files_multi_datablocks():
    dataset_model = RawDataset(
        contactEmail="p.stibbons@uu.am",
        creationTime=dateutil.parser.parse("2022-01-10T11:14:52-01:00"),
        principalInvestigator="librarian@uu.am",
        owner="PonderStibbons",
        sourceFolder=RemotePath("/hex/source91"),
        type=DatasetType.RAW,
        ownerGroup="faculty",
    )
    orig_datablock_models = [
        OrigDatablock(
            dataFileList=[
                DataFile(
                    path="file1.dat",
                    size=6123,
                )
            ],
            size=6123,
            ownerGroup="faculty",
        ),
        OrigDatablock(
            dataFileList=[
                DataFile(
                    path="sub/file2.png",
                    size=992,
                )
            ],
            size=992,
            ownerGroup="faculty",
        ),
    ]
    dset = Dataset.from_models(
        dataset_model=dataset_model,
        orig_datablock_models=orig_datablock_models,
    )

    assert len(list(dset.files)) == 2
    assert dset.number_of_files == 2
    assert dset.number_of_files_archived == 0
    assert dset.packed_size == 0
    assert dset.size == 6123 + 992

    f0 = [f for f in dset.files if f.remote_path.suffix == ".dat"][0]
    assert f0.remote_access_path(dset.source_folder) == "/hex/source91/file1.dat"
    assert f0.local_path is None
    assert f0.size == 6123
    assert f0.make_model().path == "file1.dat"

    f1 = [f for f in dset.files if f.remote_path.suffix == ".png"][0]
    assert f1.remote_access_path(dset.source_folder) == "/hex/source91/sub/file2.png"
    assert f1.local_path is None
    assert f1.size == 992
    assert f1.make_model().path == "sub/file2.png"


def test_fields_type_filter_derived():
    assert all(
        field.used_by_derived for field in Dataset.fields(dataset_type="derived")
    )


def test_fields_type_filter_raw():
    assert all(field.used_by_raw for field in Dataset.fields(dataset_type="raw"))


def test_fields_read_only_filter_true():
    assert all(field.read_only for field in Dataset.fields(read_only=True))


def test_fields_read_only_filter_false():
    assert all(not field.read_only for field in Dataset.fields(read_only=False))


def test_fields_read_only__and_type_filter():
    assert all(
        not field.read_only and field.used_by_raw
        for field in Dataset.fields(read_only=False, dataset_type="raw")
    )


def test_make_raw_model():
    dset = Dataset(
        type="raw",
        contact_email="p.stibbons@uu.am",
        creation_time="2142-04-02T16:44:56",
        owner="Ponder Stibbons;Mustrum Ridcully",
        owner_group="faculty",
        principal_investigator="my principal investigator",
        source_folder=RemotePath("/hex/source62"),
        creation_location="ANK/UU",
        shared_with=["librarian", "hicks"],
    )
    expected = RawDataset(
        contactEmail="p.stibbons@uu.am",
        creationTime=dateutil.parser.parse("2142-04-02T16:44:56"),
        owner="Ponder Stibbons;Mustrum Ridcully",
        ownerGroup="faculty",
        principalInvestigator="my principal investigator",
        sourceFolder=RemotePath("/hex/source62"),
        type=DatasetType.RAW,
        history=[],
        isPublished=False,
        scientificMetadata={},
        creationLocation="ANK/UU",
        sharedWith=["librarian", "hicks"],
        numberOfFiles=0,
        numberOfFilesArchived=0,
        packedSize=0,
        size=0,
    )
    assert dset.make_model() == expected


def test_make_derived_model():
    dset = Dataset(
        type="derived",
        contact_email="p.stibbons@uu.am;m.ridcully@uu.am",
        creation_time="2142-04-02T16:44:56",
        owner="Ponder Stibbons;Mustrum Ridcully",
        owner_group="faculty",
        investigator="p.stibbons@uu.am",
        source_folder=RemotePath("/hex/source62"),
        meta={"weight": {"value": 5.23, "unit": "kg"}},
        input_datasets=[PID(pid="623-122")],
        used_software=["scitacean", "magick"],
    )
    expected = DerivedDataset(
        contactEmail="p.stibbons@uu.am;m.ridcully@uu.am",
        creationTime=dateutil.parser.parse("2142-04-02T16:44:56"),
        owner="Ponder Stibbons;Mustrum Ridcully",
        ownerGroup="faculty",
        investigator="p.stibbons@uu.am",
        sourceFolder=RemotePath("/hex/source62"),
        type=DatasetType.DERIVED,
        history=[],
        isPublished=False,
        scientificMetadata={"weight": {"value": 5.23, "unit": "kg"}},
        inputDatasets=[PID(pid="623-122")],
        usedSoftware=["scitacean", "magick"],
        numberOfFiles=0,
        numberOfFilesArchived=0,
        packedSize=0,
        size=0,
    )
    assert dset.make_model() == expected


@pytest.mark.parametrize(
    "field",
    filter(
        lambda f: not f.used_by_raw,
        Dataset.fields(dataset_type="derived", read_only=False),
    ),
    ids=lambda f: f.name,
)
@given(st.data())
@settings(max_examples=10)
def test_make_raw_model_raises_if_derived_field_set(field, data):
    dset = Dataset(
        type="raw",
        contact_email="p.stibbons@uu.am",
        creation_time="2142-04-02T16:44:56",
        owner="Mustrum Ridcully",
        owner_group="faculty",
        investigator="p.stibbons@uu.am",
        source_folder=RemotePath("/hex/source62"),
    )
    setattr(dset, field.name, data.draw(st.from_type(field.type)))
    with pytest.raises(ValueError):
        dset.make_model()


@pytest.mark.parametrize(
    "field",
    filter(
        lambda f: not f.used_by_derived,
        Dataset.fields(dataset_type="raw", read_only=False),
    ),
    ids=lambda f: f.name,
)
@given(st.data())
@settings(max_examples=10)
def test_make_derived_model_raises_if_raw_field_set(field, data):
    dset = Dataset(
        type="derived",
        contact_email="p.stibbons@uu.am",
        creation_time="2142-04-02T16:44:56",
        owner="Ponder Stibbons",
        owner_group="faculty",
        investigator="p.stibbons@uu.am",
        source_folder=RemotePath("/hex/source62"),
        input_datasets=[PID(pid="623-122")],
        used_software=["scitacean", "magick"],
    )
    setattr(dset, field.name, data.draw(st.from_type(field.type)))
    with pytest.raises(ValueError):
        dset.make_model()


@pytest.mark.parametrize("field", ("contact_email", "owner_email"))
def test_email_validation(field):
    dset = Dataset(
        type="raw",
        contact_email="p.stibbons@uu.am",
        creation_time="2142-04-02T16:44:56",
        owner="Mustrum Ridcully",
        owner_group="faculty",
        principal_investigator="p.stibbons@uu.am",
        source_folder=RemotePath("/hex/source62"),
    )
    setattr(dset, field, "not-an-email")
    with pytest.raises(pydantic.ValidationError):
        dset.make_model()


@pytest.mark.parametrize(
    "good_orcid",
    (
        "https://orcid.org/0000-0002-3761-3201",
        "https://orcid.org/0000-0001-2345-6789",
        "https://orcid.org/0000-0003-2818-0368",
    ),
)
def test_orcid_validation_valid(good_orcid):
    dset = Dataset(
        type="raw",
        contact_email="jan-lukas.wynen@ess.eu",
        creation_time="2142-04-02T16:44:56",
        owner="Jan-Lukas Wynen",
        owner_group="ess",
        principal_investigator="jan-lukas.wynen@ess.eu",
        source_folder=RemotePath("/hex/source62"),
        orcid_of_owner=good_orcid,
    )
    assert dset.make_model().orcidOfOwner == good_orcid


@pytest.mark.parametrize(
    "bad_orcid",
    (
        "0000-0002-3761-3201",
        "https://not-orcid.eu/0000-0002-3761-3201",
        "https://orcid.org/0010-0002-3765-3201",
        "https://orcid.org/0000-0002-3761-320X",
    ),
)
def test_orcid_validation_missing_url(bad_orcid):
    dset = Dataset(
        type="raw",
        contact_email="jan-lukas.wynen@ess.eu",
        creation_time="2142-04-02T16:44:56",
        owner="Jan-Lukas Wynen",
        owner_group="ess",
        principal_investigator="jan-lukas.wynen@ess.eu",
        source_folder=RemotePath("/hex/source62"),
        orcid_of_owner=bad_orcid,
    )
    with pytest.raises(pydantic.ValidationError):
        dset.make_model()


# TODO technique



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Generate dataset fields."""

import dataclasses
from pathlib import Path
from string import Template
from typing import Dict, List

from spec import Field, Spec
from templates import BANNER, load_template
from util import quote


def _write_dataset(target: Path, dataset: str):
    with target.open("w") as f:
        f.write(BANNER + dataset)


def _format_dataset_field_spec(fields: List[Field]) -> str:
    def format_single(template: Template, field: Field) -> str:
        return template.substitute(
            name=quote(field.name),
            description=quote(field.description),
            read_only=field.read_only,
            required_by_derived=field.required,
            required_by_raw=field.required,
            type=field.type,
            used_by_derived="derived" in field.extra.get("used", ()),
            used_by_raw="raw" in field.extra.get("used", ()),
        )

    tmpl = load_template("field_spec")
    return "[\n" + "\n".join(format_single(tmpl, f) for f in fields) + "\n    ]"


def _format_dataset_field_init_args(fields: List[Field]) -> str:
    return ",\n        ".join(
        f"{f.name}: Optional[{f.type}] = None"
        for f in fields
        if not f.read_only and not f.manual
    )


def _format_dataset_field_construction(field: Field) -> str:
    n = quote(field.name)
    d = field.default
    df = field.default_factory
    if field.read_only:
        formatted = f"{n}: _apply_default(_read_only.get({n}), {d}, {df})"
    else:
        formatted = f"{n}: _apply_default({field.name}, {d}, {df})"
    return "            " + formatted + ","


def _format_dataset_field_dict_construction(fields: List[Field]) -> str:
    return "\n".join(
        _format_dataset_field_construction(field)
        for field in fields
        if not field.manual
    )


def _format_properties(field: Field) -> str:
    getter = f'''    @property
    def {field.name}(self) -> Optional[{field.type}]:
        """{field.description}"""
        return self._fields[{quote(field.name)}]  # type: ignore[no-any-return]'''
    if field.read_only:
        return getter
    setter = f"""    @{field.name}.setter
    def {field.name}(self, val: Optional[{field.type}]) -> None:
        self._fields[{quote(field.name)}] = val"""
    return getter + "\n\n" + setter


def _format_make_model(typ: str, fields: List[Field]) -> str:
    extra_names = "\n".join(
        " " * 12 + quote(field.name) + ","
        for field in fields
        if typ not in field.extra.get("used", ())
    )
    extra_fields = f"""    extra_fields = {{
        name: None
        for name in (
{extra_names}
        )
        if getattr(self, name, None) is not None
    }}"""
    construction = "\n        ".join(
        f"{field.model_name}=self.{field.name},"
        for field in fields
        if typ in field.extra.get("used", ())
    )
    formatted = f"""def _make_{typ}_model(self) -> {typ.capitalize()}Dataset:
{extra_fields}
    return {("Derived" if typ == "derived" else "Raw")}Dataset(
        {construction}
        **extra_fields,
    )"""
    return "    " + formatted.replace("\n", "\n    ")


def _format_fields_from_model(typ: str, fields: List[Field]) -> str:
    def format_assignments(read_only, indent):
        return ("\n" + " " * indent).join(
            f"{field.name}=model.{field.model_name},"
            for field in fields
            if typ in field.extra.get("used", ())
            and field.read_only == read_only
            and not field.manual
        )

    return f"""def _fields_from_{typ}_model(model) -> dict:
    return dict(
        _read_only=dict(
            {format_assignments(read_only=True, indent=12)}
        ),
        {format_assignments(read_only=False, indent=8)}
    )
"""


def _format_dataset_dataclass(spec: Spec) -> str:
    template = load_template("dataset")
    fields = sorted(spec.fields, key=lambda field: field.name)
    properties = "\n\n".join(
        _format_properties(field) for field in fields if not field.manual
    )
    return template.substitute(
        field_spec=_format_dataset_field_spec(fields),
        field_init_args=_format_dataset_field_init_args(fields),
        field_dict_construction=_format_dataset_field_dict_construction(fields),
        properties=properties,
        make_derived_model=_format_make_model("derived", fields),
        make_raw_model=_format_make_model("raw", fields),
        fields_from_derived_model=_format_fields_from_model("derived", fields),
        fields_from_raw_model=_format_fields_from_model("raw", fields),
    )


def _inline_base(spec: Spec, specs: Dict[str, Spec]) -> Spec:
    if spec.inherits == "BaseModel":
        return spec

    base = specs[spec.inherits]
    new_spec = Spec(
        name=spec.name, inherits="BaseModel", fields=spec.fields + base.fields
    )
    if base.inherits != "BaseModel":
        new_spec.inherits = base.inherits
        return _inline_base(new_spec, specs)
    return new_spec


def _add_used_extra(spec: Spec) -> Spec:
    """Make sure that every field has extra["used"]."""
    return dataclasses.replace(
        spec,
        fields=[
            dataclasses.replace(
                field, extra={"used": ["derived", "raw"], **field.extra}
            )
            for field in spec.fields
        ],
    )


def generate_dataset_fields(target: Path, specs: Dict[str, Spec]):
    _write_dataset(
        target,
        _format_dataset_dataclass(
            _add_used_extra(_inline_base(specs["Dataset"], specs))
        ),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Filesystem utilities.

Local paths are stored as pathlib.Path
Remote paths are stored as scitacean.filesystem.RemotePath
"""

from __future__ import annotations

import hashlib
import os
import re
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Callable, Generator, Optional, TypeVar, Union


class RemotePath(os.PathLike):  # type: ignore[type-arg]
    """A path on the remote filesystem.

    Remote paths do not need to correspond to a regular filesystem path like
    :class:`pathlib.PosixPath` or :class:`pathlib.WindowsPath`.
    Instead, they can be any sequence of segments that are joined by forward slashes,
    e.g. a URL.
    """

    def __init__(self, path: Union[str, RemotePath]):
        """Initialize from a given path."""
        self._path = os.fspath(path)

    def __truediv__(self, other: Union[str, RemotePath]) -> RemotePath:
        """Join two path segments."""
        this = _strip_trailing_slash(os.fspath(self))
        other = _strip_leading_slash(_strip_trailing_slash(os.fspath(other)))
        return RemotePath(f"{this}/{other}")

    def __rtruediv__(self, other: str) -> RemotePath:
        """Join two path segments."""
        return RemotePath(other) / self

    def __str__(self) -> str:
        return self._path

    def __repr__(self) -> str:
        return f"RemotePath({str(self)})"

    def __fspath__(self) -> str:
        """Return the file system representation of the path."""
        return str(self)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (RemotePath, str)):
            return False
        return self._path == RemotePath(other)._path

    def __hash__(self) -> int:
        return hash(self._path)

    @property
    def name(self) -> str:
        """The name of the file with all directories removed."""
        return self._path.rstrip("/").rsplit("/", 1)[-1]

    @property
    def suffix(self) -> Optional[str]:
        """The file extension including a leading period."""
        parts = self.name.rsplit(".", 1)
        if len(parts) == 1:
            return None
        return "." + parts[1]

    def truncated(self, max_length: int = 255) -> RemotePath:
        """Return a new remote path with all path segments truncated.

        Parameters
        ----------
        max_length:
            Maximum length of each segment.
            The default value is the typical maximum length on Linux.

        Returns
        -------
        :
            A new remote path with truncated segments.
        """

        def trunc(seg: str) -> str:
            # First, make sure that the name is short enough to fit the suffix
            # such that the suffix does not get truncated.
            # But keep at least one character of the name.
            # Then make sure the whole segment is short enough, potentially
            # truncating the suffix.
            parts = seg.rsplit(".", 1)
            name = parts[0]
            suffix = "." + parts[1] if len(parts) > 1 else ""
            return (name[: max(1, max_length - len(suffix))] + suffix)[:max_length]

        return RemotePath("/".join(map(trunc, self._path.split("/"))))

    @classmethod
    def __get_validators__(
        cls,
    ) -> Generator[Callable[[Union[str, RemotePath]], RemotePath], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, value: Union[str, RemotePath]) -> RemotePath:
        """Pydantic validator for RemotePath fields."""
        return RemotePath(value)


def _strip_trailing_slash(s: str) -> str:
    return s[:-1] if s.endswith("/") else s


def _strip_leading_slash(s: str) -> str:
    return s[1:] if s.startswith("/") else s


def file_size(path: Path) -> int:
    """Return the size of a local file in bytes."""
    return path.stat().st_size


def file_modification_time(path: Path) -> datetime:
    """Return the time in UTC when a local file was last modified."""
    return datetime.fromtimestamp(path.stat().st_mtime).astimezone(timezone.utc)


def _new_hash(algorithm: str) -> Any:
    try:
        return hashlib.new(algorithm, usedforsecurity=False)
    except TypeError:
        # Fallback for Python < 3.9
        return hashlib.new(algorithm)


# size based on http://git.savannah.gnu.org/gitweb/?p=coreutils.git;a=blob;f=src/ioblksize.h;h=ed2f4a9c4d77462f357353eb73ee4306c28b37f1;hb=HEAD#l23  # noqa: E501
def checksum_of_file(path: Union[str, Path], *, algorithm: str) -> str:
    """Compute the checksum of a local file.

    Parameters
    ----------
    path:
        Path of the file on the local filesystem.
    algorithm:
        Hash algorithm to use. Can be any algorithm supported by :func:`hashlib.new`.

    Returns
    -------
    :
        The hex digest of the hash.
    """
    chk = _new_hash(algorithm)
    buffer = memoryview(bytearray(128 * 1024))
    with open(path, "rb", buffering=0) as file:
        for n in iter(lambda: file.readinto(buffer), 0):
            chk.update(buffer[:n])
    return chk.hexdigest()  # type: ignore[no-any-return]


P = TypeVar("P", bound=Union[str, Path, RemotePath])


def escape_path(path: P) -> P:
    """Escape disallowed characters for file paths.

    Replaces

    - Unicode characters using ``"backslashreplace"``.
      See the `Python docs <https://docs.python.org/3/library/codecs.html?highlight=unicode_escape#error-handlers>`_.
    - Non-word characters by '_'. This includes backslashes
      introduced by the above.

    The result should be a valid path name on Linux, macOS, and Windows.

    Parameters
    ----------
    path:
        Input string or path.

    Returns
    -------
    :
        ``path`` with offending characters replaced.
        Has the same type as ``path``.
    """  # noqa: E501
    s = os.fspath(path)
    no_utf = s.encode("ascii", "backslashreplace").decode("ascii")
    return type(path)(re.sub(r"[^\w .\-]", "_", no_utf))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Generate pydantic models and DatasetFields."""

import argparse
from pathlib import Path

from dataset import generate_dataset_fields
from model import generate_models
from spec import load_specs

# TODO model validation shows model names (aliases?)


def parse_args():
    parser = argparse.ArgumentParser("Generate models and dataclasses")
    parser.add_argument("--src-dir", type=Path, help="Scitacean source directory")
    return parser.parse_args()


def main():
    args = parse_args()
    specs = load_specs()
    generate_models(args.src_dir / "model.py", specs)
    generate_dataset_fields(args.src_dir / "_dataset_fields.py", specs)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Fake client for testing."""

from __future__ import annotations

import datetime
import functools
import uuid
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Union

from .. import model
from ..client import Client, ScicatClient
from ..error import ScicatCommError
from ..pid import PID
from ..typing import FileTransfer
from ..util.credentials import StrStorage


def _conditionally_disabled(func: Callable[..., Any]) -> Callable[..., Any]:
    @functools.wraps(func)
    def impl(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        if (exc := self.main.disabled.get(func.__name__)) is not None:
            raise exc
        return func(self, *args, **kwargs)

    return impl


# Inherits from Client to satisfy type hints.
class FakeClient(Client):
    """Mimics a client without accessing the internet.

    This class is a stand in for :class:`scitacean.Client` to emulate downloading and
    uploading of datasets without actually accessing a SciCat server.

    Since this client fakes communication with SciCat, it stores raw
    `pydantic <https://docs.pydantic.dev/>`_ models, namely

    - ``FakeClient.datasets``:
            :class:`dict` of :class:`pydantic.model.DerivedDataset` and
            :class:`pydantic.model.RawDataset`,
            indexed by dataset ID.
    - ``FakeClient.orig_datablocks``:
            :class:`dict` of lists of :class:`pydantic.model.OrigDatablock`,
            indexed by the *dataset* ID.

    Individual functions can be disabled (that is made to raise an exception)
    using the ``disabled`` argument of the initializer.

    Important
    ---------
    ``FakeClient`` does not behave exactly like a ``Client`` connected to a real
    server as that would require reimplementing a large part of the SciCat backend.
    You should thus always test your code against a (potentially locally deployed)
    real server.

    In particular, do not rely on specific error messages or the detailed settings
    in the datasets returned by ``FakeClient.upload_new_dataset_now``!

    Examples
    --------
    Set up a fake client for download:

    .. code-block:: python

        client = FakeClient()
        pid = PID(prefix="sample.prefix", pid="1234-5678-abcd")
        client.datasets[pid] = model.DerivedDataset(...)
        client.orig_datablocks[pid] = [model.OrigDatablock(
            datasetId=str(pid),
            ...
        )]
        client.get_dataset(pid)

    Use to upload a dataset:

    .. code-block:: python

        client = FakeClient(file_transfer=FakeFileTransfer())
        finalized = client.upload_new_dataset_now(dset)

        # contains new dataset and datablock:
        client.datasets[finalized.pid]
        client.orig_datablocks[finalized.pid]

    Disable a method:

    .. code-block:: python

        client = FakeClient(
            disable={"create_dataset_model": RuntimeError("Upload failed")})
        # raises RuntimeError("Upload failed"):
        client.upload_new_dataset_now(dset)

    See Also
    --------
    scitacean.testing.transfer.FakeFileTransfer: Fake file up-/downloads
        without a real server.
    """

    def __init__(
        self,
        *,
        file_transfer: Optional[FileTransfer] = None,
        disable: Optional[Dict[str, Exception]] = None,
    ) -> None:
        """Initialize a fake client with empty dataset storage.

        Parameters
        ----------
        file_transfer:
            Typically :class:`scitacean.testing.transfer.FakeFileTransfer`
            but may be a real file transfer.
        disable:
            dict function names to exceptions. Functions listed here raise
            the given exception when called and do nothing else.
        """
        super().__init__(client=FakeScicatClient(self), file_transfer=file_transfer)

        self.disabled = {} if disable is None else dict(disable)
        self.datasets: Dict[PID, Union[model.DerivedDataset, model.RawDataset]] = {}
        self.orig_datablocks: Dict[PID, List[model.OrigDatablock]] = {}

    @classmethod
    def from_token(
        cls,
        *,
        url: str,
        token: Union[str, StrStorage],
        file_transfer: Optional[FileTransfer] = None,
    ) -> FakeClient:
        """Create a new fake client.

        All arguments except ``file_Transfer`` are ignored.
        """
        return FakeClient(file_transfer=file_transfer)

    @classmethod
    def from_credentials(
        cls,
        *,
        url: str,
        username: Union[str, StrStorage],
        password: Union[str, StrStorage],
        file_transfer: Optional[FileTransfer] = None,
    ) -> FakeClient:
        """Create a new fake client.

        All arguments except ``file_Transfer`` are ignored.
        """
        return FakeClient(file_transfer=file_transfer)

    @classmethod
    def without_login(
        cls, *, url: str, file_transfer: Optional[FileTransfer] = None
    ) -> FakeClient:
        """Create a new fake client.

        All arguments except ``file_Transfer`` are ignored.
        """
        return FakeClient(file_transfer=file_transfer)


class FakeScicatClient(ScicatClient):
    """Mimics a ScicatClient, to be used by FakeClient."""

    def __init__(self, main_client: FakeClient) -> None:
        super().__init__(url="", token="", timeout=datetime.timedelta(seconds=60))
        self.main = main_client

    @_conditionally_disabled
    def get_dataset_model(
        self, pid: PID, strict_validation: bool = False
    ) -> Union[model.DerivedDataset, model.RawDataset]:
        _ = strict_validation  # unused by fake
        try:
            return self.main.datasets[pid]
        except KeyError:
            raise ScicatCommError(f"Unable to retrieve dataset {pid}") from None

    @_conditionally_disabled
    def get_orig_datablocks(
        self, pid: PID, strict_validation: bool = False
    ) -> List[model.OrigDatablock]:
        _ = strict_validation  # unused by fake
        try:
            return self.main.orig_datablocks[pid]
        except KeyError:
            raise ScicatCommError(
                f"Unable to retrieve orig datablock for dataset {pid}"
            ) from None

    @_conditionally_disabled
    def create_dataset_model(
        self, dset: Union[model.DerivedDataset, model.RawDataset]
    ) -> Union[model.DerivedDataset, model.RawDataset]:
        pid = PID(
            pid=str(dset.pid) if dset.pid is not None else str(uuid.uuid4()),
            prefix="PID.SAMPLE.PREFIX",
        )
        if pid in self.main.datasets:
            raise ScicatCommError(f"Dataset id already exists: {pid}")
        self.main.datasets[pid] = deepcopy(dset)
        self.main.datasets[pid].pid = pid
        return self.main.datasets[pid]

    @_conditionally_disabled
    def create_orig_datablock(self, dblock: model.OrigDatablock) -> model.OrigDatablock:
        dataset_id = dblock.datasetId
        if dataset_id not in self.main.datasets:
            raise ScicatCommError(f"No dataset with id {dataset_id}")
        self.main.orig_datablocks.setdefault(dataset_id, []).append(dblock)
        return dblock



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Tools for securely handling credentials."""
from __future__ import annotations

import datetime
from typing import NoReturn, Optional, Union


class StrStorage:
    """Base class for storing a string.

    Instances can be nested to combine different specialized features.
    """

    def __init__(self, value: Optional[Union[str, StrStorage]]):
        self._value = value

    def get_str(self) -> str:
        """Return the stored plain str object."""
        if self._value is None:
            # If the implementation chooses to not
            # store the string in memory.
            # The method must be overridden in this case.
            raise NotImplementedError("String not available")

        if isinstance(self._value, StrStorage):
            return self._value.get_str()
        return self._value

    def __str__(self) -> str:
        return str(self._value)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._value!r})"


# TODO implement
# class KeyringStr(StrStorage):
#     """Store a string in the user's keyring.
#
#     Should be the bottom level StrStorage because it erases any nested
#     StrStorage objects and stores and returns plain strs.
#     """
#
#     def __init__(self, *, key: str, value: Optional[Union[str, StrStorage]]):
#         super().__init__(None)
#         # TODO dummy implementation
#         self._ring = {}
#         self._key = key
#         if value is not None:
#             if isinstance(value, StrStorage):
#                 self._store(value.get_str())
#             else:
#                 self._store(value)
#
#     def _store(self, value: str):
#         self._ring[self._key] = value
#
#     def _retrieve(self) -> str:
#         return self._ring[self._key]
#
#     def get_str(self) -> str:
#         return self._retrieve()
#
#     def __str__(self) -> str:
#         return "???"
#
#     def __repr__(self) -> str:
#         return f"KeyringStr(key='{self._key}', value={str(self)})"


class SecretStr(StrStorage):
    """Minimize the risk of exposing a secret.

    Stores a string and blocks the most common pathways
    of leaking it to logs or files.

    Attention
    ---------
    The string may be stored as a regular, unencrypted str object and can
    still be leaked through introspection methods.
    """

    def __init__(self, value: Union[str, StrStorage]):
        super().__init__(value)

    def __str__(self) -> str:
        return "***"

    def __repr__(self) -> str:
        return "SecretStr(***)"

    # prevent pickling
    def __reduce_ex__(self, protocol: object) -> NoReturn:
        raise TypeError("SecretStr must not be pickled")


class TimeLimitedStr(StrStorage):
    """A string that expires after some time."""

    def __init__(
        self,
        *,
        value: Union[str, StrStorage],
        expires_at: datetime.datetime,
        tolerance: Optional[datetime.timedelta] = None,
    ):
        super().__init__(value)
        if tolerance is None:
            tolerance = datetime.timedelta(seconds=10)
        self._expires_at = expires_at - tolerance

    def get_str(self) -> str:
        if self._is_expired():
            raise RuntimeError("Login has expired")
        return super().get_str()

    def _is_expired(self) -> bool:
        return datetime.datetime.now() > self._expires_at

    def __repr__(self) -> str:
        return (
            f"TimeLimitedStr(expires_at={self._expires_at.isoformat()}, "
            f"value={repr(self._value)}"
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

from contextlib import contextmanager

import pytest
from dateutil.parser import parse as parse_date

from scitacean import Dataset, DatasetType, ScicatCommError, model
from scitacean.testing.client import FakeClient
from scitacean.testing.transfer import FakeFileTransfer

from .common.files import make_file

# TODO source_folder changed / populated in file after upload


@pytest.fixture
def ownable():
    return model.Ownable(ownerGroup="uu", accessGroups=["group1", "2nd_group"])


@pytest.fixture
def derived_dataset_model(ownable):
    return model.DerivedDataset(
        owner="PonderStibbons",
        investigator="ridcully@uu.am",
        contactEmail="p.stibbons@uu.am",
        sourceFolder="/hex/source123",
        size=0,
        numberOfFiles=0,
        creationTime=parse_date("2011-08-24T12:34:56Z"),
        datasetName="Data A38",
        inputDatasets=[],
        usedSoftware=["EasyScience"],
        scientificMetadata={
            "data_type": "event data",
            "temperature": {"value": "123", "unit": "K"},
            "weight": {"value": "42", "unit": "mg"},
        },
        type=DatasetType.DERIVED,
        **ownable.dict(),
    )


@pytest.fixture
def client(fs):
    return FakeClient.from_token(
        url="",
        token="",
        file_transfer=FakeFileTransfer(fs=fs, files={}, reverted={}),
    )


@pytest.fixture
def dataset(derived_dataset_model, fs):
    make_file(fs, path="file.nxs", contents=b"contents of file.nxs")
    make_file(fs, path="the_log_file.log", contents=b"this is a log file")
    dset = Dataset.from_models(
        dataset_model=derived_dataset_model, orig_datablock_models=None
    )
    dset.add_local_files("file.nxs", "the_log_file.log")
    return dset


def test_upload_returns_updated_dataset(client, dataset):
    finalized = client.upload_new_dataset_now(dataset)
    expected = client.get_dataset(finalized.pid).replace(
        # The backend may update the dataset after upload
        _read_only={
            "updated_at": finalized.updated_at,
            "updated_by": finalized.updated_by,
        }
    )
    assert finalized == expected


def test_upload_without_files_creates_dataset(client, derived_dataset_model):
    dataset = Dataset.from_models(
        dataset_model=derived_dataset_model, orig_datablock_models=None
    )
    finalized = client.upload_new_dataset_now(dataset)
    expected = client.get_dataset(finalized.pid).replace(
        # The backend may update the dataset after upload
        _read_only={
            "updated_at": finalized.updated_at,
            "updated_by": finalized.updated_by,
        }
    )
    assert finalized == expected
    with pytest.raises(ScicatCommError):
        client.scicat.get_orig_datablocks(finalized.pid)


def test_upload_uses_given_pid(client, derived_dataset_model):
    derived_dataset_model.pid = "my-custom-id"
    dataset = Dataset.from_models(
        dataset_model=derived_dataset_model, orig_datablock_models=None
    )
    finalized = client.upload_new_dataset_now(dataset)
    assert "my-custom-id" in str(finalized.pid)


def test_upload_creates_dataset_and_datablock(client, dataset):
    finalized = client.upload_new_dataset_now(dataset)
    assert client.datasets[finalized.pid] == finalized.make_model()
    assert (
        client.orig_datablocks[finalized.pid]
        == finalized.make_datablock_models().orig_datablocks
    )


def test_upload_uploads_files_to_source_folder(client, dataset):
    finalized = client.upload_new_dataset_now(dataset)
    source_folder = client.file_transfer.source_folder_for(finalized)

    assert (
        client.file_transfer.files[source_folder / "file.nxs"]
        == b"contents of file.nxs"
    )
    assert (
        client.file_transfer.files[source_folder / "the_log_file.log"]
        == b"this is a log file"
    )


def test_upload_does_not_create_dataset_if_file_upload_fails(dataset, fs):
    class RaisingUpload(FakeFileTransfer):
        source_dir = "/"

        def upload_files(self, *files):
            raise RuntimeError("Fake upload failure")

        @contextmanager
        def connect_for_upload(self, pid):
            yield self

    client = FakeClient(file_transfer=RaisingUpload(fs=fs))

    with pytest.raises(RuntimeError, match="Fake upload failure"):
        client.upload_new_dataset_now(dataset)

    assert dataset.pid not in client.datasets
    assert dataset.pid not in client.orig_datablocks


def test_upload_cleans_up_files_if_dataset_ingestion_fails(dataset, fs):
    client = FakeClient(
        disable={"create_dataset_model": ScicatCommError("Ingestion failed")},
        file_transfer=FakeFileTransfer(fs=fs),
    )
    with pytest.raises(ScicatCommError):
        client.upload_new_dataset_now(dataset)

    assert not client.file_transfer.files


def test_failed_datablock_upload_does_not_revert(dataset, fs):
    client = FakeClient(
        disable={"create_orig_datablock": ScicatCommError("Ingestion failed")},
        file_transfer=FakeFileTransfer(fs=fs),
    )
    with pytest.raises(RuntimeError):
        client.upload_new_dataset_now(dataset)

    uploaded_dset = next(iter(client.datasets.values()))
    assert uploaded_dset.owner == "PonderStibbons"
    assert uploaded_dset.datasetName == "Data A38"
    assert uploaded_dset.usedSoftware == ["EasyScience"]

    assert client.file_transfer.files
    assert not client.file_transfer.reverted



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

"""Wrappers for (Orig)Datablocks."""

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Iterator, List, Optional, Union

from .file import File
from .model import DerivedDataset, OrigDatablock, RawDataset
from .pid import PID

if TYPE_CHECKING:
    from .dataset import Dataset

# TODO DatablockProxy


@dataclasses.dataclass
class OrigDatablockProxy:
    """Dataclass for an orig datablock.

    Instances of this class are mutable as opposed to
    :class:`scitacean.model.OrigDatablock`.
    They are used for building datasets and get converted to/from pydantic
    models for communication with a server.
    """

    _files: List[File] = dataclasses.field(init=False)
    _files_modified: bool = dataclasses.field(default=False, init=False)
    checksum_algorithm: Optional[str] = None
    pid: Optional[PID] = None
    owner_group: Optional[str] = None
    access_groups: Optional[List[str]] = None
    instrument_group: Optional[str] = None
    init_files: dataclasses.InitVar[Optional[List[File]]] = None

    def __post_init__(self, init_files: Optional[List[File]]) -> None:
        self._files = list(init_files) if init_files is not None else []

    @classmethod
    def from_model(
        cls,
        *,
        dataset_model: Union[DerivedDataset, RawDataset],
        orig_datablock_model: OrigDatablock,
    ) -> OrigDatablockProxy:
        """Construct a new OrigDatablockProxy from pydantic models.

        Parameters
        ----------
        dataset_model:
            Model of the dataset that this orig datablock belongs to.
        orig_datablock_model:
            Model of the orig datablock to construct.

        Returns
        -------
        :
            A new instance.
        """
        dblock = orig_datablock_model
        # TODO store checksum once implemented
        #   AND overwrite in Files
        return OrigDatablockProxy(
            pid=dblock.id,
            checksum_algorithm=None,
            owner_group=dblock.ownerGroup,
            access_groups=dblock.accessGroups,
            instrument_group=dblock.instrumentGroup,
            init_files=[File.from_scicat(file) for file in dblock.dataFileList],
        )

    @property
    def files(self) -> Iterator[File]:
        """Iterator over all files."""
        return iter(self._files)

    @property
    def size(self) -> int:
        """Total size of all files."""
        return sum(file.size for file in self.files)

    def add_files(self, *files: File) -> None:
        """Append files to the datablock.

        Parameters
        ----------
        files:
            File objects to add.
        """
        self._files.extend(
            dataclasses.replace(f, checksum_algorithm=self.checksum_algorithm)
            for f in files
        )
        self._files_modified = True

    def make_model(self, dataset: Dataset) -> OrigDatablock:
        """Build a new pydantic model for this datablock.

        Parameters
        ----------
        dataset:
            The dataset that this orig datablock belongs to.

        Returns
        -------
        :
            A new model for this orig datablock.
        """
        # TODO set checksum_algorithm once implemented
        return OrigDatablock(
            id=self.pid,
            size=self.size,
            dataFileList=[file.make_model(for_archive=False) for file in self.files],
            datasetId=dataset.pid,
            ownerGroup=self.owner_group or dataset.owner_group,
            accessGroups=self.access_groups or dataset.access_groups,
            instrumentGroup=self.instrument_group or dataset.instrument_group,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
import json
import shutil
import tempfile
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Union
from urllib.parse import urljoin

import pytest
import requests

from ..data import load_datasets, load_orig_datablocks
from .docker import docker_compose

_TEST_BASE = Path(__file__).resolve().parent.parent
_SCICAT_DOCKER_CONFIG = _TEST_BASE / "scicatlive/docker-compose.yaml"
_SCICAT_DATASET_SEED_FILE = Path("seed_db/seed/Dataset.json")
_SCICAT_ORIG_DATABLOCK_SEED_FILE = Path("seed_db/seed/OrigDatablock.json")


# List of required services for tests.
# We only need the backend and API to run tests.
_SERVICES = ("catamel", "mongodb", "mongodb_seed", "reverse-proxy")


@dataclass
class SciCatAccess:
    url: str
    functional_credentials: Dict[str, str]


@pytest.fixture(scope="session")
def scicat_access():
    return SciCatAccess(
        url="http://localhost/api/v3/",
        functional_credentials={"username": "ingestor", "password": "aman"},
    )


@pytest.fixture(scope="session")
def scicat_backend(request, scicat_access):
    """Spin up a SciCat backend and API.

    Does nothing unless the --backend-tests command line option is set.
    """
    if not request.config.getoption("--backend-tests"):
        yield False
        return

    with tempfile.TemporaryDirectory() as temp_dir:
        config_file = configure(temp_dir)
        with docker_compose(config_file, *_SERVICES):
            wait_until_backend_is_live(scicat_access, max_time=20, n_tries=20)
            yield True


def can_connect(scicat_access: SciCatAccess) -> bool:
    response = requests.post(
        urljoin(scicat_access.url, "Users/login"),
        json=scicat_access.functional_credentials,
        timeout=0.5,
    )
    return response.ok


def wait_until_backend_is_live(
    scicat_access: SciCatAccess, max_time: float, n_tries: int
):
    # The containers take a while to be fully live.
    for _ in range(n_tries):
        if can_connect(scicat_access):
            return
        time.sleep(max_time / n_tries)
    if not can_connect(scicat_access):
        raise RuntimeError("Cannot connect to backend")


def _merge_seed_file(
    target_dir: Path, seed_file_name: Union[str, Path], custom_seed: Union[list, dict]
):
    with open(target_dir / seed_file_name, "r") as f:
        dset_seed = json.load(f)
    dset_seed.extend(custom_seed)
    with open(target_dir / seed_file_name, "w") as f:
        json.dump(dset_seed, f)


def configure(target_dir: Path) -> Path:
    target = Path(target_dir) / "scicatlive"
    shutil.copytree(
        _SCICAT_DOCKER_CONFIG.parent, target, ignore=shutil.ignore_patterns(".git")
    )
    _merge_seed_file(target, _SCICAT_DATASET_SEED_FILE, load_datasets())
    _merge_seed_file(target, _SCICAT_ORIG_DATABLOCK_SEED_FILE, load_orig_datablocks())

    return target / _SCICAT_DOCKER_CONFIG.name


def skip_if_not_backend(request):
    if not request.config.getoption("--backend-tests"):
        # The backend only exists if this option is set.
        pytest.skip(
            "Tests against a real backend are disabled, "
            "use --backend-tests to enable them"
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

from scitacean import Dataset, RemotePath


# We don't want to test the concrete layout as that may change
# without breaking anything. So just make sure that the result
# contains the relevant data.
def test_dataset_html_repr():
    ds = Dataset(
        type="raw",
        name="My dataset",
        contact_email="devsci.cat",
        owner="The People",
        instrument_id="the-peoples-neutron-gun",
        used_software=["scitacean"],
        source_folder=RemotePath("/remote/dir/"),
        meta={
            "temperature": {"value": 5, "unit": "C"},
            "mood": "has been better",
        },
    )

    res = ds._repr_html_()

    assert "My dataset" in res
    assert "used_software" in res
    assert "temperature" in res
    assert "unit" in res
